"""Release logic for the GitHub API."""
import json

from uritemplate import URITemplate

from .. import models
from .. import users
from .. import utils
from ..decorators import requires_auth
from ..exceptions import error_for


class Release(models.GitHubCore):
    """Representation of a GitHub release.

    It holds the information GitHub returns about a release from a
    :class:`Repository <github3.repos.repo.Repository>`.

    Please see GitHub's `Releases Documentation`_ for more information.

    This object has the following attributes:

    .. attribute:: original_assets

        A list of :class:`~github3.repos.release.Asset` objects representing
        the assets uploaded for this relesae.

    .. attribute:: assets_url

        The URL to retrieve the assets from the API.

    .. attribute:: author

        A :class:`~github3.users.ShortUser` representing the creator of this
        release.

    .. attribute:: body

        The description of this release as written by the release creator.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        when this release was created.

    .. attribute:: draft

        A boolean attribute describing whether this release is a draft.

    .. attribute:: html_url

        The URL to view this release in a browser.

    .. attribute:: id

        The unique identifier of this release.

    .. attribute:: name

        The name given to this release by the :attr:`author`.

    .. attribute:: prerelease

        A boolean attribute indicating whether the release is a pre-release.

    .. attribute:: published_at

        A :class:`~datetime.datetime` object representing the date and time
        when this release was publisehd.

    .. attribute:: tag_name

        The name of the tag associated with this release.

    .. attribute:: tarball_url

        The URL to retrieve a GitHub generated tarball for this release from
        the API.

    .. attribute:: target_commitish

        The reference (usually a commit) that is targetted by this release.

    .. attribute:: upload_urlt

        A :class:`~uritemplate.URITemplate` object that expands to form the
        URL to upload assets to.

    .. attribute:: zipball_url

        The URL to retrieve a GitHub generated zipball for this release from
        the API.

    .. _Releases Documentation:
        https://developer.github.com/v3/repos/releases/
    """

    def _update_attributes(self, release):
        self._api = self.url = release["url"]
        self.original_assets = [Asset(i, self) for i in release["assets"]]
        self.assets_url = release["assets_url"]
        self.author = users.ShortUser(release["author"], self)
        self.body = release["body"]
        self.created_at = self._strptime(release["created_at"])
        self.draft = release["draft"]
        self.html_url = release["html_url"]
        self.id = release["id"]
        self.name = release["name"]
        self.prerelease = release["prerelease"]
        self.published_at = self._strptime(release["published_at"])
        self.tag_name = release["tag_name"]
        self.tarball_url = release["tarball_url"]
        self.target_commitish = release["target_commitish"]
        self.upload_urlt = URITemplate(release["upload_url"])
        self.zipball_url = release["zipball_url"]

    def _repr(self):
        return f"<Release [{self.name}]>"

    def archive(self, format, path=""):
        """Get the tarball or zipball archive for this release.

        :param str format:
            (required), accepted values: ('tarball', 'zipball')
        :param path:
            (optional), path where the file should be saved to, default is the
            filename provided in the headers and will be written in the current
            directory. It can take a file-like object as well
        :type path:
            str, file
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        resp = None
        if format in ("tarball", "zipball"):
            repo_url = self._api[: self._api.rfind("/releases")]
            url = self._build_url(format, self.tag_name, base_url=repo_url)
            resp = self._get(url, allow_redirects=True, stream=True)

        if resp and self._boolean(resp, 200, 404):
            utils.stream_response_to_file(resp, path)
            return True
        return False

    def asset(self, asset_id):
        """Retrieve the asset from this release with ``asset_id``.

        :param int asset_id:
            ID of the Asset to retrieve
        :returns:
            the specified asset, if it exists
        :rtype:
            :class:`~github3.repos.release.Asset`
        """
        json = None
        if int(asset_id) > 0:
            i = self._api.rfind("/")
            url = self._build_url(
                "assets", str(asset_id), base_url=self._api[:i]
            )
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Asset, json)

    def assets(self, number=-1, etag=None):
        """Iterate over the assets available for this release.

        :param int number:
            (optional), Number of assets to return
        :param str etag:
            (optional), last ETag header sent
        :returns:
            generator of asset objects
        :rtype:
            :class:`~github3.repos.release.Asset`
        """
        url = self._build_url("assets", base_url=self._api)
        return self._iter(number, url, Asset, etag=etag)

    @requires_auth
    def delete(self):
        """Delete this release.

        Only users with push access to the repository can delete a release.

        :returns:
            True if successful; False if not successful
        :rtype:
            bool
        """
        url = self._api
        return self._boolean(self._delete(url), 204, 404)

    @requires_auth
    def edit(
        self,
        tag_name=None,
        target_commitish=None,
        name=None,
        body=None,
        draft=None,
        prerelease=None,
    ):
        """Edit this release.

        Only users with push access to the repository can edit a release.

        If the edit is successful, this object will update itself.

        :param str tag_name:
            (optional), Name of the tag to use
        :param str target_commitish:
            (optional), The "commitish" value that determines where the Git tag
            is created from. Defaults to the repository's default branch.
        :param str name:
            (optional), Name of the release
        :param str body:
            (optional), Description of the release
        :param boolean draft:
            (optional), True => Release is a draft
        :param boolean prerelease:
            (optional), True => Release is a prerelease
        :returns:
            True if successful; False if not successful
        :rtype:
            bool
        """
        url = self._api
        data = {
            "tag_name": tag_name,
            "target_commitish": target_commitish,
            "name": name,
            "body": body,
            "draft": draft,
            "prerelease": prerelease,
        }
        self._remove_none(data)

        r = self.session.patch(url, data=json.dumps(data))

        successful = self._boolean(r, 200, 404)
        if successful:
            # If the edit was successful, let's update the object.
            self._update_attributes(r.json())

        return successful

    @requires_auth
    def upload_asset(self, content_type, name, asset, label=None):
        """Upload an asset to this release.

        .. note:: All parameters are required.

        :param str content_type:
            The content type of the asset. Wikipedia has a list of common media
            types
        :param str name:
            The name of the file
        :param asset:
            The file or bytes object to upload.
        :param label:
            (optional), An alternate short description of the asset.
        :returns:
            the created asset
        :rtype:
            :class:`~github3.repos.release.Asset`
        """
        headers = {"Content-Type": content_type}
        params = {"name": name, "label": label}
        self._remove_none(params)
        url = self.upload_urlt.expand(params)
        r = self._post(url, data=asset, json=False, headers=headers)
        if r.status_code in (201, 202):
            return Asset(r.json(), self)
        raise error_for(r)


class Asset(models.GitHubCore):
    """Representation of an asset in a release.

    .. seealso::

        `List Assets`_, :meth:`~github3.repos.release.Release.assets`
            Documentation around listing assets of a release
        `Upload Assets`_, :meth:`~github3.repos.release.Release.upload_asset`
            Documentation around uploading assets to a release
        `Get a Single Asset`_, :meth:`~github3.repos.release.Release.asset`
            Documentation around retrieving an asset
        `Edit an Asset`_, :meth:`~github3.repos.release.Asset.edit`
            Documentation around editing an asset
        `Delete an Asset`_, :meth:`~github3.repos.release.Asset.delete`
            Documentation around deleting an asset

    This object has the following attributes:

    .. attribute:: browser_download_url

        The user-friendly URL to download this asset via a browser.

    .. attribute:: content_type

        The Content-Type provided by the uploader when this asset was created.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        when this asset was created.

    .. attribute:: download_count

        The number of times this asset has been downloaded.

    .. attribute:: download_url

        The URL to retrieve this uploaded asset via the API, e.g., tarball,
        zipball, etc.

    .. attribute:: id

        The unique identifier of this asset.

    .. attribute:: label

        The short description of this asset.

    .. attribute:: name

        The name provided to this asset.

    .. attribute:: size

        The file size of this asset.

    .. attribute:: state

        The state of this asset, e.g., ``'uploaded'``.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        when this asset was most recently updated.

    .. _List Assets:
        https://developer.github.com/v3/repos/releases/#list-assets-for-a-release
    .. _Upload Assets:
        https://developer.github.com/v3/repos/releases/#upload-a-release-asset
    .. _Get a Single Asset:
        https://developer.github.com/v3/repos/releases/#get-a-single-release-asset
    .. _Edit an Asset:
        https://developer.github.com/v3/repos/releases/#edit-a-release-asset
    .. _Delete an Asset:
        https://developer.github.com/v3/repos/releases/#delete-a-release-asset
    """

    def _update_attributes(self, asset):
        self._api = asset["url"]
        self.browser_download_url = asset["browser_download_url"]
        self.content_type = asset["content_type"]
        self.created_at = self._strptime(asset["created_at"])
        self.download_count = asset["download_count"]
        self.download_url = self._api
        self.id = asset["id"]
        self.label = asset["label"]
        self.name = asset["name"]
        self.size = asset["size"]
        self.state = asset["state"]
        self.updated_at = self._strptime(asset["updated_at"])

    def _repr(self):
        return f"<Asset [{self.name}]>"

    def download(self, path=""):
        """Download the data for this asset.

        :param path:
            (optional), path where the file should be saved to, default is the
            filename provided in the headers and will be written in the current
            directory. It can take a file-like object as well
        :type path:
            str, file
        :returns:
            name of the file, if successful otherwise ``None``
        :rtype:
            str
        """
        headers = {"Accept": "application/octet-stream"}
        resp = self._get(
            self._api, allow_redirects=False, stream=True, headers=headers
        )
        if resp.status_code == 302:
            # Amazon S3 will reject the redirected request unless we omit
            # certain request headers
            headers.update({"Content-Type": None})

            with self.session.no_auth():
                resp = self._get(
                    resp.headers["location"], stream=True, headers=headers
                )

        if self._boolean(resp, 200, 404):
            return utils.stream_response_to_file(resp, path)
        return None

    @requires_auth
    def delete(self):
        """Delete this asset if the user has push access.

        :returns:
            True if successful; False if not successful
        :rtype:
            bool
        """
        url = self._api
        return self._boolean(self._delete(url), 204, 404)

    def edit(self, name, label=None):
        """Edit this asset.

        :param str name:
            (required), The file name of the asset
        :param str label:
            (optional), An alternate description of the asset
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        if not name:
            return False
        edit_data = {"name": name, "label": label}
        self._remove_none(edit_data)
        r = self._patch(self._api, data=json.dumps(edit_data))
        successful = self._boolean(r, 200, 404)
        if successful:
            self._update_attributes(r.json())

        return successful



"""This module provides decorators to the rest of the library."""
import os
from functools import wraps
from io import BytesIO as StringIO

from requests.models import Response


class RequestsStringIO(StringIO):
    """Shim compatibility for string IO."""

    def read(self, n=-1, *args, **kwargs):
        """Ignore extra args and kwargs."""
        # StringIO is an old-style class, so can't use super
        return StringIO.read(self, n)


def requires_auth(func):
    """Decorator to note which object methods require authorization."""

    @wraps(func)
    def auth_wrapper(self, *args, **kwargs):
        if hasattr(self, "session") and self.session.has_auth():
            return func(self, *args, **kwargs)
        else:
            from .exceptions import error_for

            # Mock a 401 response
            r = generate_fake_error_response(
                '{"message": "Requires authentication"}'
            )
            raise error_for(r)

    return auth_wrapper


def requires_basic_auth(func):
    """Specific (basic) authentication decorator.

    This is used to note which object methods require username/password
    authorization and won't work with token based authorization.

    """

    @wraps(func)
    def auth_wrapper(self, *args, **kwargs):
        if hasattr(self, "session") and self.session.auth:
            return func(self, *args, **kwargs)
        else:
            from .exceptions import error_for

            # Mock a 401 response
            r = generate_fake_error_response(
                '{"message": "Requires username/password authentication"}'
            )
            raise error_for(r)

    return auth_wrapper


def requires_app_credentials(func):
    """Require client_id and client_secret to be associated.

    This is used to note and enforce which methods require a client_id and
    client_secret to be used.

    """

    @wraps(func)
    def auth_wrapper(self, *args, **kwargs):
        client_id, client_secret = self.session.retrieve_client_credentials()
        if client_id and client_secret:
            return func(self, *args, **kwargs)
        else:
            from .exceptions import error_for

            # Mock a 401 response
            r = generate_fake_error_response(
                '{"message": "Requires username/password authentication"}'
            )
            raise error_for(r)

    return auth_wrapper


def requires_app_bearer_auth(func):
    """Require the use of application authentication.

    .. versionadded:: 1.2.0
    """

    @wraps(func)
    def auth_wrapper(self, *args, **kwargs):
        from . import session

        if isinstance(self.session.auth, session.AppBearerTokenAuth):
            return func(self, *args, **kwargs)
        else:
            from . import exceptions

            raise exceptions.MissingAppBearerAuthentication(
                "This method requires GitHub App authentication."
            )

    return auth_wrapper


def requires_app_installation_auth(func):
    """Require the use of App's installation authentication.

    .. versionadded:: 1.2.0
    """

    @wraps(func)
    def auth_wrapper(self, *args, **kwargs):
        from . import session

        if isinstance(self.session.auth, session.AppInstallationTokenAuth):
            return func(self, *args, **kwargs)
        else:
            from . import exceptions

            raise exceptions.MissingAppInstallationAuthentication(
                "This method requires GitHub App authentication."
            )

    return auth_wrapper


def generate_fake_error_response(msg, status_code=401, encoding="utf-8"):
    """Generate a fake Response from requests."""
    r = Response()
    r.status_code = status_code
    r.encoding = encoding
    r.raw = RequestsStringIO(msg.encode())
    r._content_consumed = True
    r._content = r.raw.read()
    return r


# Use mock decorators when generating documentation, so all functino signatures
# are displayed correctly
if os.getenv("GENERATING_DOCUMENTATION", None) == "github3":
    requires_auth = requires_basic_auth = lambda x: x  # noqa  # (No coverage)



"""Module containing the GistHistory object."""
from .. import models
from .. import users


class GistHistory(models.GitHubCore):
    """This object represents one version (or revision) of a gist.

    The GitHub API returns the following attributes:

    .. attribute:: url

        The URL to the revision of the gist retrievable through the API.

    .. attribute:: version

        The commit ID of the revision of the gist.

    .. attribute:: user

        The :class:`~github3.users.ShortUser` representation of the user who
        owns this gist.

    .. attribute:: committed_at

        The date and time of the revision's commit.

    .. attribute:: change_status

        A dictionary with the number of deletions, additions, and total
        changes to the gist.

    For convenience, github3.py also exposes the following attributes from the
    :attr:`change_status`:

    .. attribute:: additions

        The number of additions to the gist compared to the previous revision.

    .. attribute:: deletions

        The number of deletions from the gist compared to the previous
        revision.

    .. attribute:: total

        The total number of changes to the gist compared to the previous
        revision.
    """

    def _update_attributes(self, history) -> None:
        self.url = self._api = history["url"]
        self.version = history["version"]
        self.user = users.ShortUser(history["user"], self)
        self.change_status = history["change_status"]
        self.additions = self.change_status.get("additions")
        self.deletions = self.change_status.get("deletions")
        self.total = self.change_status["total"]
        self.committed_at = self._strptime(history["committed_at"])

    def _repr(self) -> str:
        return f"<Gist History [{self.version}]>"

    def gist(self):
        """Retrieve the gist at this version.

        :returns:
            the gist at this point in history or ``None``
        :rtype:
            :class:`Gist <github3.gists.gist.Gist>`
        """
        from .gist import Gist

        json = self._json(self._get(self._api), 200)
        return self._instance_or_null(Gist, json)



"""Code search results implementation."""
from .. import models
from .. import repos


class CodeSearchResult(models.GitHubCore):
    """A representation of a code search result from the API.

    This object has the following attributes:

    .. attribute:: git_url

        The URL to retrieve the blob via Git

    .. attribute:: html_url

        The URL to view the blob found in a browser.

    .. attribute:: name

        The name of the file where the search result was found.

    .. attribute:: path

        The path in the repository to the file containing the result.

    .. attribute:: repository

        A :class:`~github3.repos.repo.ShortRepository` representing the
        repository in which the result was found.

    .. attribute:: score

        The confidence score assigned to the result.

    .. attribute:: sha

        The SHA1 of the blob in which the code can be found.

    .. attribute:: text_matches

        A list of the text matches in the blob that generated this result.

        .. note::

            To receive these, you must pass ``text_match=True`` to
            :meth:`~github3.github.GitHub.search_code`.
    """

    def _update_attributes(self, data):
        self._api = data["url"]
        self.git_url = data["git_url"]
        self.html_url = data["html_url"]
        self.name = data["name"]
        self.path = data["path"]
        self.repository = repos.ShortRepository(data["repository"], self)
        self.score = data["score"]
        self.sha = data["sha"]
        self.text_matches = data.get("text_matches", [])

    def _repr(self):
        return f"<CodeSearchResult [{self.path}]>"



"""Module containing helpers for dealing with GitHub Apps.

https://developer.github.com/apps/building-github-apps/
"""
import time

import jwt

from . import models
from . import users

TEN_MINUTES_AS_SECONDS = 10 * 60
DEFAULT_JWT_TOKEN_EXPIRATION = TEN_MINUTES_AS_SECONDS
APP_PREVIEW_HEADERS = {
    "Accept": "application/vnd.github.machine-man-preview+json"
}


class App(models.GitHubCore):
    """An object representing a GitHub App.

    .. versionadded:: 1.2.0

    .. seealso::

        `GitHub Apps`_
            Documentation for Apps on GitHub

        `GitHub Apps API Documentation`_
            API documentation of what's available about an App.

    This object has the following attributes:

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the day and time
        the App was created.

    .. attribute:: description

        The description of the App provided by the owner.

    .. attribute:: events

        An array of the event types an App receives

    .. attribute:: external_url

        The URL provided for the App by the owner.

    .. attribute:: html_url

        The HTML URL provided for the App by the owner.

    .. attribute:: id

        The unique identifier for the App. This is useful in cases where you
        may want to authenticate either as an App or as a specific
        installation of an App.

    .. attribute:: name

        The display name of the App that the user sees.

    .. attribute:: node_id

        A base64-encoded blob returned by the GitHub API for who knows what
        reason.

    .. attribute:: owner

        A :class:`~github3.users.ShortUser` object representing the GitHub
        user who owns the App.

    .. attribute:: permissions

        A dictionary describing the permissions the App has

    .. attribute:: slug

        A short string used to identify the App

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the day and time
        the App was last updated.

    .. _GitHub Apps:
        https://developer.github.com/apps/
    .. _GitHub Apps API Documentation:
        https://developer.github.com/v3/apps/
    """

    CUSTOM_HEADERS = {
        "Accept": "application/vnd.github.machine-man-preview+json"
    }

    def _update_attributes(self, json):
        self.created_at = self._strptime(json["created_at"])
        self.description = json["description"]
        self.external_url = json["external_url"]
        self.events = json["events"]
        self.html_url = json["html_url"]
        self.id = json["id"]
        self.name = json["name"]
        self.node_id = json["node_id"]
        self.owner = users.ShortUser(json["owner"], self)
        self.permissions = json["permissions"]
        self.slug = json["slug"]
        self.updated_at = self._strptime(json["updated_at"])
        self._api = self.url = self._build_url("apps", self.slug)

    def _repr(self):
        return f'<App ["{self.name}" by {str(self.owner)}]>'


class Installation(models.GitHubCore):
    """An installation of a GitHub App either on a User or Org.

    .. versionadded:: 1.2.0

    This has the following attributes:

    .. attribute:: access_tokens_url
    .. attribute:: account
    .. attribute:: app_id
    .. attribute:: created_at
    .. attribute:: events
    .. attribute:: html_url
    .. attribute:: id
    .. attribute:: permissions
    .. attribute:: repositories_url
    .. attribute:: repository_selection
    .. attribute:: single_file_name
    .. attribute:: target_id
    .. attribute:: target_type
    .. attribute:: updated_at
    """

    def _update_attributes(self, json):
        self.access_tokens_url = json["access_tokens_url"]
        self.account = json["account"]
        self.app_id = json["app_id"]
        self.created_at = self._strptime(json["created_at"])
        self.events = json["events"]
        self.html_url = json["html_url"]
        self.id = json["id"]
        self.permissions = json["permissions"]
        self.repositories_url = json["repositories_url"]
        self.repository_selection = json["repository_selection"]
        self.single_file_name = json["single_file_name"]
        self.target_id = json["target_id"]
        self.target_type = json["target_type"]
        self.updated_at = self._strptime(json["updated_at"])


def create_token(private_key_pem, app_id, expire_in=TEN_MINUTES_AS_SECONDS):
    """Create an encrypted token for the specified App.

    :param bytes private_key_pem:
        The bytes of the private key for this GitHub Application.
    :param int app_id:
        The integer identifier for this GitHub Application.
    :param int expire_in:
        The length in seconds for this token to be valid for.
        Default: 600 seconds (10 minutes)
    :returns:
        Serialized encrypted token.
    :rtype:
        text
    """
    if not isinstance(private_key_pem, bytes):
        raise ValueError('"private_key_pem" parameter must be byte-string')
    now = int(time.time())
    token = jwt.encode(
        payload={"iat": now, "exp": now + expire_in, "iss": app_id},
        key=private_key_pem,
        algorithm="RS256",
    )
    return token


def create_jwt_headers(
    private_key_pem, app_id, expire_in=DEFAULT_JWT_TOKEN_EXPIRATION
):
    """Create an encrypted token for the specified App.

    :param bytes private_key_pem:
        The bytes of the private key for this GitHub Application.
    :param int app_id:
        The integer identifier for this GitHub Application.
    :param int expire_in:
        The length in seconds for this token to be valid for.
        Default: 600 seconds (10 minutes)
    :returns:
        Dictionary of headers for retrieving a token from a JWT.
    :rtype:
        dict
    """
    jwt_token = create_token(private_key_pem, app_id, expire_in)
    headers = {"Authorization": f"Bearer {jwt_token}"}
    headers.update(APP_PREVIEW_HEADERS)
    return headers



"""This module provides the basic models used in github3.py."""
import json as jsonlib
import logging
import typing as t

import dateutil.parser
import requests.compat

from . import exceptions
from . import session


if t.TYPE_CHECKING:
    from . import structs

LOG = logging.getLogger(__package__)


T = t.TypeVar("T")


class GitHubCore:
    """The base object for all objects that require a session.

    The :class:`GitHubCore <GitHubCore>` object provides some
    basic attributes and methods to other sub-classes that are very useful to
    have.
    """

    _ratelimit_resource = "core"
    _refresh_to: t.Optional["GitHubCore"] = None

    def __init__(self, json, session: session.GitHubSession):
        """Initialize our basic object.

        Pretty much every object will pass in decoded JSON and a Session.
        """
        # Either or 'session' is an instance of a GitHubCore sub-class or it
        # is a session. In the former case it will have a 'session' attribute.
        # If it doesn't, we can just default to using the passed in session.
        self.session = getattr(session, "session", session)

        # set a sane default
        self._github_url = "https://api.github.com"

        if json is not None:
            self.etag = json.pop("ETag", None)
            self.last_modified = json.pop("Last-Modified", None)
            self._uniq = json.get("url", None)
        self._json_data = json
        try:
            self._update_attributes(json)
        except KeyError as kerr:
            raise exceptions.IncompleteResponse(json, kerr)

    def _update_attributes(self, json):
        pass

    def __getattr__(self, attribute):
        """Proxy access to stored JSON."""
        _json_data = object.__getattribute__(self, "_json_data")
        if attribute not in _json_data:
            raise AttributeError(attribute)
        value = _json_data[attribute]
        setattr(self, attribute, value)
        return value

    def as_dict(self):
        """Return the attributes for this object as a dictionary.

        This is equivalent to calling::

            json.loads(obj.as_json())

        :returns: this object's attributes serialized to a dictionary
        :rtype: dict
        """
        return self._json_data

    def as_json(self):
        """Return the json data for this object.

        This is equivalent to calling::

            json.dumps(obj.as_dict())

        :returns: this object's attributes as a JSON string
        :rtype: str
        """
        return jsonlib.dumps(self._json_data)

    @classmethod
    def _strptime(cls, time_str):
        """Convert an ISO 8601 formatted string to a datetime object.

        We assume that the ISO 8601 formatted string is in UTC and we create
        the datetime object so that it is timezone-aware.

        :param str time_str: ISO 8601 formatted string
        :returns: timezone-aware datetime object
        :rtype: datetime or None
        """
        if time_str:
            return dateutil.parser.parse(time_str)
        return None

    def __repr__(self):
        repr_string = self._repr()
        if requests.compat.is_py2:
            return repr_string.encode("utf-8")
        return repr_string

    @classmethod
    def from_dict(cls, json_dict, session):
        """Return an instance of this class formed from ``json_dict``."""
        return cls(json_dict, session)

    @classmethod
    def from_json(cls, json, session):
        """Return an instance of this class formed from ``json``."""
        return cls(jsonlib.loads(json), session)

    def __eq__(self, other):
        return self._uniq == other._uniq

    def __ne__(self, other):
        return self._uniq != other._uniq

    def __hash__(self):
        return hash(self._uniq)

    def _repr(self):
        return f"<github3-core at 0x{id(self):x}>"

    @staticmethod
    def _remove_none(data):
        if not data:
            return
        for k, v in list(data.items()):
            if v is None:
                del data[k]

    def _instance_or_null(self, instance_class, json):
        if json is not None and not isinstance(json, dict):
            raise exceptions.UnprocessableResponseBody(
                "GitHub's API returned a body that could not be handled", json
            )
        if not json:
            return None

        return instance_class(json, self)

    def _json(self, response, expected_status_code, include_cache_info=True):
        ret = None
        if response is None:
            return None

        actual_status_code = response.status_code
        if actual_status_code != expected_status_code:
            if actual_status_code >= 400:
                raise exceptions.error_for(response)

            if actual_status_code == 304:
                # Received a response from someone passing in `etag=`
                return None

            LOG.warning(
                "Expected status_code %d but got %d",
                expected_status_code,
                actual_status_code,
            )

        try:
            ret = response.json()
        except ValueError:
            raise exceptions.UnexpectedResponse(response)

        headers = response.headers
        if (
            include_cache_info
            and (headers.get("Last-Modified") or headers.get("ETag"))
            and isinstance(ret, dict)
        ):
            ret["Last-Modified"] = response.headers.get("Last-Modified", "")
            ret["ETag"] = response.headers.get("ETag", "")
        LOG.info("JSON was %sreturned", "not " if ret is None else "")
        return ret

    def _boolean(self, response, true_code, false_code):
        if response is not None:
            status_code = response.status_code
            if status_code == true_code:
                return True
            if status_code != false_code and status_code >= 400:
                raise exceptions.error_for(response)
        return False

    def _request(self, method, *args, **kwargs):
        try:
            request_method = getattr(self.session, method)
            return request_method(*args, **kwargs)
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.Timeout,
        ) as exc:
            raise exceptions.ConnectionError(exc)
        except requests.exceptions.RequestException as exc:
            raise exceptions.TransportError(exc)

    def _delete(self, url, **kwargs):
        LOG.debug("DELETE %s with %s", url, kwargs)
        return self._request("delete", url, **kwargs)

    def _get(self, url, **kwargs):
        LOG.debug("GET %s with %s", url, kwargs)
        return self._request("get", url, **kwargs)

    def _patch(self, url, **kwargs):
        LOG.debug("PATCH %s with %s", url, kwargs)
        return self._request("patch", url, **kwargs)

    def _post(self, url, data=None, json=True, **kwargs):
        if json:
            data = jsonlib.dumps(data) if data is not None else data
        LOG.debug("POST %s with %s, %s", url, data, kwargs)
        return self._request("post", url, data, **kwargs)

    def _put(self, url, **kwargs):
        LOG.debug("PUT %s with %s", url, kwargs)
        return self._request("put", url, **kwargs)

    def _build_url(self, *args, **kwargs):
        """Build a new API url from scratch."""
        return self.session.build_url(*args, **kwargs)

    @property
    def _api(self):
        value = "{0.scheme}://{0.netloc}{0.path}".format(self._uri)
        if self._uri.query:
            value += f"?{self._uri.query}"
        return value

    @staticmethod
    def _uri_parse(uri):
        return requests.compat.urlparse(uri)

    @_api.setter
    def _api(self, uri):
        if uri:
            self._uri = self._uri_parse(uri)
        self.url = uri

    def _iter(
        self,
        count: int,
        url: str,
        cls: t.Type[T],
        params: t.Optional[t.Mapping[str, t.Optional[str]]] = None,
        etag: t.Optional[str] = None,
        headers: t.Optional[t.Mapping[str, str]] = None,
        list_key: t.Optional[str] = None,
    ) -> "structs.GitHubIterator[T]":
        """Generic iterator for this project.

        :param int count: How many items to return.
        :param int url: First URL to start with
        :param class cls: cls to return an object of
        :param params dict: (optional) Parameters for the request
        :param str etag: (optional), ETag from the last call
        :param dict headers: (optional) HTTP Headers for the request
        :param str list_key: (optional) Key for extracting the list of items
            from a dict response
        :returns: A lazy iterator over the pagianted resource
        :rtype: :class:`GitHubIterator <github3.structs.GitHubIterator>`
        """
        from .structs import GitHubIterator

        return GitHubIterator(
            count, url, cls, self, params, etag, headers, list_key
        )

    @property
    def ratelimit_remaining(self):
        """Number of requests before GitHub imposes a ratelimit.

        :returns: int
        """
        json = self._json(self._get(self._build_url("rate_limit")), 200)
        core = json.get("resources", {}).get(self._ratelimit_resource, {})
        self._remaining = core.get("remaining", 0)
        return self._remaining

    def refresh(self, conditional: bool = False) -> "GitHubCore":
        """Re-retrieve the information for this object.

        The reasoning for the return value is the following example: ::

            repos = [r.refresh() for r in g.repositories_by('kennethreitz')]

        Without the return value, that would be an array of ``None``'s and you
        would otherwise have to do: ::

            repos = [r for i in g.repositories_by('kennethreitz')]
            [r.refresh() for r in repos]

        Which is really an anti-pattern.

        .. versionchanged:: 0.5

        .. _Conditional Requests:
            http://developer.github.com/v3/#conditional-requests

        :param bool conditional: If True, then we will search for a stored
            header ('Last-Modified', or 'ETag') on the object and send that
            as described in the `Conditional Requests`_ section of the docs
        :returns: self
        """
        headers = getattr(self, "CUSTOM_HEADERS", {})
        if conditional:
            if self.last_modified:
                headers["If-Modified-Since"] = self.last_modified
            elif self.etag:
                headers["If-None-Match"] = self.etag

        headers = headers or None
        json = self._json(self._get(self._api, headers=headers), 200)
        if json is not None:
            if self._refresh_to is None:
                self._json_data = json
                self._update_attributes(json)
            else:
                return self._refresh_to(json, self)
        return self

    def new_session(self):
        """Generate a new session.

        :returns:
            A brand new session
        :rtype:
            :class:`~github3.session.GitHubSession`
        """
        return session.GitHubSession()



"""This module contains all the classes relating to Checks."""
from json import dumps

from . import decorators
from . import models


class CheckPullRequest(models.GitHubCore):
    """Representation of a Pull Request returned in Checks APIs.

    .. versionadded:: 1.3.0

    .. note::

        Refreshing this object returns a :class:`~github3.pulls.PullRequest`.

    This object has the following attributes:

    .. attribute:: id

        The unique id of this pull request across all of GitHub.

    .. attribute:: number

        The number of this pull request on its repository.

    .. attribute:: head

        A dict of minimal head information retrieved from the Check data
        representing the source of the pull request

    .. attribute:: base

        A dict of minimal base information retrieved from the Check data
        representing the pull request destination.
    """

    def _update_attributes(self, pull):
        self.id = pull["id"]
        self.number = pull["number"]
        self.base = pull["base"]
        self.head = pull["head"]
        self._api = self.url = pull["url"]

    def _repr(self):
        return f"<CheckPullRequest [#{self.number}]>"

    def to_pull(self):
        """Retrieve a full PullRequest object for this CheckPullRequest.

        :returns:
            The full information about this pull request.
        :rtype:
            :class:`~github3.pulls.PullRequest`
        """
        from . import pulls

        json = self._json(self._get(self.url), 200)
        return self._instance_or_null(pulls.PullRequest, json)

    refresh = to_pull


class CheckApp(models.GitHubCore):
    """Representation of an App returned in Checks APIs.

    .. versionadded:: 1.3.0

    .. note::

        Refreshing this object returns a :class:`~github3.apps.App`.

    This object has the following attributes:

    .. attribute:: description

        The description of the App provided by the owner.

    .. attribute:: external_url

        The URL provided for the App by the owner.

    .. attribute:: html_url

        The HTML URL provided for the App by the owner.

    .. attribute:: id

        The unique identifier for the App. This is useful in cases where you
        may want to authenticate either as an App or as a specific
        installation of an App.

    .. attribute:: name

        The display name of the App that the user sees.

    .. attribute:: owner

        A dict of minimal user information retrieved from the Check data
        representing the app owner
    """

    CUSTOM_HEADERS = {
        "Accept": "application/vnd.github.machine-man-preview+json"
    }

    def _update_attributes(self, app):
        self.description = app["description"]
        self.external_url = app["external_url"]
        self.html_url = app["html_url"]
        self.id = app["id"]
        self.name = app["name"]
        self.owner = app["owner"]
        _, slug = app["html_url"].rsplit("/", 1)
        self._api = self.url = self._build_url("apps", slug)

    def _repr(self):
        return '<App ["{}" by {}]>'.format(
            self.name, str(self.owner["login"])
        )

    def to_app(self):
        """Retrieve a full App object for this CheckApp.

        :returns:
            The full information about this App.
        :rtype:
            :class:`~github3.apps.App`
        """
        from . import apps

        headers = getattr(self, "CUSTOM_HEADERS", None)
        json = self._json(self._get(self._api, headers=headers), 200)
        return self._instance_or_null(apps.App, json)

    refresh = to_app


class CheckSuite(models.GitHubCore):
    """The :class:`CheckSuite <CheckSuite>` object.

    .. versionadded:: 1.3.0

    Please see GitHub's `CheckSuite Documentation`_ for more information.

    .. attribute:: status

        The status of the Check Suite

    .. attribute:: conclusion

        The highest priority check run conclusion. If it has not completed this
        will be None

    .. attribute:: head_sha

        The sha of the commit at the head of the branch the check was run
        against (the source of the pull request)

    .. attribute:: head_branch

        The branch checked

    .. attribute:: before

        The sha of the pull request target branch at the time of the checks

    .. attribute:: after

        The sha of the target branch after the change is applied

    .. attribute:: repository

        A representation of the repository the suite belongs to as
        :class:`~github3.repos.repo.ShortRepository`.

    .. attribute:: original_pull_requests

        A list of representations of the pull requests the suite belongs to as
        :class:`~github3.checks.CheckPullRequest`.

        .. note::

            This may be empty.

    .. attribute:: id

        The unique GitHub assigned numerical id of this check suite.

    .. attribute:: app

        A :class:`~github3.checks.CheckApp` representing the App
        this suite belongs to.

    .. CheckSuite Documentation:
        http://developer.github.com/v3/checks/suites/
    """

    class_name = "CheckSuite"
    CUSTOM_HEADERS = {"Accept": "application/vnd.github.antiope-preview+json"}

    def _update_attributes(self, suite):
        # Import here, because a toplevel import causes an import loop
        from . import repos

        self._api = suite["url"]
        self.status = suite["status"]
        self.conclusion = suite["conclusion"]
        self.head_branch = suite["head_branch"]
        self.head_sha = suite["head_sha"]
        self.before = suite["before"]
        self.after = suite["after"]
        prs = suite.get("pull_requests", [])
        self.original_pull_requests = [CheckPullRequest(p, self) for p in prs]
        self.repository = repos.ShortRepository(suite["repository"], self)
        self.id = suite["id"]
        self.app = CheckApp(suite["app"], self)

    def _repr(self):
        return "<{s.class_name} [{s.id}:{s.status}]>".format(s=self)

    @decorators.requires_auth
    def rerequest(self):
        """Rerequest the check suite.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("rerequest", base_url=self._api)
        return self._boolean(
            self._post(url, headers=CheckSuite.CUSTOM_HEADERS), 201, 404
        )

    def check_runs(self):
        """Retrieve the check runs for this suite.

        :returns:
            the check runs for this commit
        :rtype:
            :class:`~github3.checks.CheckRun`
        """
        url = self._build_url("check-runs", base_url=self._api)
        return self._iter(
            -1,
            url,
            CheckRun,
            headers=CheckRun.CUSTOM_HEADERS,
            list_key="check_runs",
        )


class CheckRunAnnotation(models.GitHubCore):
    """Representation of an annotation of a check run.

    .. versionadded:: 1.3.0

    .. attribute:: path

        The path of the file of the annotation.

    .. attribute:: start_line

        The start line of the annotation.

    .. attribute:: end_line

        The end line of the annotation.

    .. attribute:: start_column

        The start column of the annotation.

    .. attribute:: end_column

        The end column of the annotation.

    .. attribute:: annotation_level

        The level of the annotation. Can be one of 'notice', 'warning'
        or 'failure'.

    .. attribute:: title

        The title that represents the annotation.

    .. attribute:: message

        The short descript of the feedback for the lines of code.

    .. attribute:: raw_details

        The details about this annotation.
    """

    class_name = "CheckRunAnnotation"
    CUSTOM_HEADERS = {"Accept": "application/vnd.github.antiope-preview+json"}

    def _repr(self):
        return (
            "<{s.class_name} [{s.path}:{s.start_line}-{s.end_line}]>"
        ).format(s=self)

    def _update_attributes(self, note):
        self.path = note["path"]
        self.start_line = note["start_line"]
        self.end_line = note["end_line"]
        self.start_column = note["start_column"]
        self.end_column = note["end_column"]
        self.annotation_level = note["annotation_level"]
        self.title = note["title"]
        self.message = note["message"]
        self.raw_details = note["raw_details"]


class CheckRunOutput(models.GitHubCore):
    """Representation of the output of a check run.

    .. versionadded:: 1.3.0

    .. attribute:: title

        The title for the check run output.

    .. attribute:: summary

        The summary for the check run output.

    .. attribute:: text

        The output text of the check run.

    .. attribute:: annotations_count

        The number of annotations for this check run.

    .. attribute:: annotations_url

        The URL to retrieve the annotations
    """

    class_name = "CheckRunOutput"

    def _update_attributes(self, output):
        self.title = output["title"]
        self.summary = output["summary"]
        self.text = output["text"]
        self.annotations_count = output["annotations_count"]
        self.annotations_url = output["annotations_url"]

    def _repr(self):
        return "<{s.class_name} [{s.title}]>".format(s=self)

    def annotations(self):
        """Retrieve the annotations for a check run.

        :returns:
            the annotations for this check run
        :rtype:
            :class:`~github3.checks.CheckRunAnnotations`
        """
        url = self._build_url(base_url=self.annotations_url)
        return self._iter(
            -1,
            url,
            CheckRunAnnotation,
            headers=CheckRunAnnotation.CUSTOM_HEADERS,
        )


class CheckRun(models.GitHubCore):
    """The :class:`CheckRun <CheckRun>` object.

    .. versionadded:: 1.3.0

    Please see GitHub's `CheckRun Documentation`_ for more information.

    .. attribute:: status

        The current status of the check.

    .. attribute:: conclusion

        The final conclusion of the check. If the run has not concluded
        this will be None.

    .. attribute:: head_sha

        The sha of the commit at the head of the branch checked.

    .. attribute:: name

        The name of the check.

    .. attribute:: started_at

        A :class:`~datetime.datetime` object representing the date and time
        when this check run started.

    .. attribute:: completed_at

        A :class:`~datetime.datetime` object representing the date and time
        when this check run completed. If this run is not completed it will
        be ``None``.

    .. attribute:: original_pull_requests

        A list of representations of the pull requests the run belongs to as
        :class:`~github3.checks.CheckPullRequest`.

        .. note::

            This may be empty.


    .. attribute:: id

        The unique GitHub assigned numerical id of this check run.

    .. attribute:: external_id

        A reference for the run on the integrator's system. This may be None.

    .. attribute:: html_url

        The URL one would use to view this check run in the browser.

    .. attribute:: check_suite

        The ID of the check suite this run belongs to.

    .. attribute:: output

        A :class:`~github3.checks.CheckRunOutput` representing the output
        of this check run.

    .. attribute:: app

        A :class:`~github3.checks.CheckApp` representing the App
        this run belongs to.

    .. CheckRun Documentation:
        http://developer.github.com/v3/checks/runs/
    """

    class_name = "CheckRun"
    CUSTOM_HEADERS = {"Accept": "application/vnd.github.antiope-preview+json"}

    def _update_attributes(self, run):
        self._api = run["url"]
        self.html_url = run["html_url"]
        self.status = run["status"]
        self.conclusion = run["conclusion"]
        self.started_at = self._strptime(run["started_at"])
        self.completed_at = self._strptime(run["completed_at"])
        self.head_sha = run["head_sha"]
        self.name = run["name"]
        prs = run.get("pull_requests", [])
        self.original_pull_requests = [CheckPullRequest(p, self) for p in prs]
        self.id = run["id"]
        self.external_id = run["external_id"]
        self.app = CheckApp(run["app"], self)
        self.check_suite = run["check_suite"]["id"]
        self.output = CheckRunOutput(run["output"], self)

    def _repr(self):
        return "<{s.class_name} [{s.name}:{s.status}]>".format(s=self)

    @decorators.requires_app_installation_auth
    def update(
        self,
        name=None,
        details_url=None,
        external_id=None,
        started_at=None,
        status=None,
        conclusion=None,
        completed_at=None,
        output=None,
        actions=None,
    ):
        """Update this check run.

        All parameters are optional.

        :param str name:
            (optional), new name of the check
        :param str details_url:
            (optional), The URL of the integrator's site that has the full
            details of the check
        :param str external_id:
            (optional), A reference for the run on the integrator's system
        :param str started_at:
            (optional), ISO 8601 time format: YYYY-MM-DDTHH:MM:SSZ
        :param str status:
            (optional), ('queued', 'in_progress', 'completed')
        :param str conclusion:
            (optional), Required if you provide 'completed_at', or a
            'status' of 'completed'. The final conclusion of the check.
            ('success', 'failure', 'neutral', 'cancelled', 'timed_out',
            'action_required')
        :param str completed_at:
            (optional), Required if you provide 'conclusion'. ISO 8601 time
            format: YYYY-MM-DDTHH:MM:SSZ
        :param dict output:
            (optional), key-value pairs representing the output. Format:
            ``{'title': 'string', 'summary', 'text, can be markdown', 'text':
            'text, can be markdown', 'annotations': [{}], 'images': [{}]}``
        :param list actions:
            (optional), list of action objects. Format is:
            ``[{'label': 'text', 'description', 'text', 'identifier', 'text'},
            ...]``
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        data = {
            "name": name,
            "details_url": details_url,
            "external_id": external_id,
            "started_at": started_at,
            "status": status,
            "conclusion": conclusion,
            "completed_at": completed_at,
            "output": output,
            "actions": actions,
        }
        self._remove_none(data)
        json = None

        if data:
            json = self._json(
                self._patch(
                    self._api,
                    data=dumps(data),
                    headers=CheckSuite.CUSTOM_HEADERS,
                ),
                200,
            )
        if json:
            self._update_attributes(json)
            return True
        return False



"""Commit search results implementation."""
from .. import git
from .. import models
from .. import repos
from .. import users


class CommitSearchResult(models.GitHubCore):
    """A representation of a commit search result from the API.

    This object has the following attributes:

    .. attribute:: author

        A :class:`~github3.users.ShortUser` representing the user who
        authored the found commit.

    .. attribute:: comments_url

        The URL to retrieve the comments on the found commit from the API.

    .. attribute:: commit

        A :class:`~github3.git.ShortCommit` representing the found commit.

    .. attribute:: committer

        A :class:`~github3.users.ShortUser` representing the user who
        committed the found commit.

    .. attribute:: html_url

        The URL to view the found commit in a browser.

    .. attribute:: repository

        A :class:`~github3.repos.repo.ShortRepository` representing the
        repository in which the commit was found.

    .. attribute:: score

        The confidence score assigned to the result.

    .. attribute:: sha

        The SHA1 of the found commit.

    .. attribute:: text_matches

        A list of the text matches in the commit that generated this result.

        .. note::

            To receive these, you must pass ``text_match=True`` to
            :meth:`~github3.github.GitHub.search_commit`.
    """

    def _update_attributes(self, data):
        self._api = data["url"]
        self.author = users.ShortUser(data["author"], self)
        self.comments_url = data["comments_url"]
        self.commit = git.ShortCommit(data["commit"], self)
        self.committer = users.ShortUser(data["committer"], self)
        self.html_url = data["html_url"]
        self.repository = repos.ShortRepository(data["repository"], self)
        self.score = data["score"]
        self.sha = data["sha"]
        self.text_matches = data.get("text_matches", [])

    def _repr(self):
        return f"<CommitSearchResult [{self.sha[:7]}]>"



import collections.abc
import functools
import typing as t

from requests.compat import urlencode
from requests.compat import urlparse

from . import exceptions
from . import models

if t.TYPE_CHECKING:
    import requests.models

    from . import session


T = t.TypeVar("T")


class GitHubIterator(models.GitHubCore, collections.abc.Iterator):
    """The :class:`GitHubIterator` class powers all of the iter_* methods."""

    def __init__(
        self,
        count: int,
        url: str,
        cls: t.Type[T],
        session: "session.GitHubSession",
        params: t.Optional[t.Mapping[str, t.Optional[str]]] = None,
        etag: t.Optional[str] = None,
        headers: t.Optional[t.Mapping[str, str]] = None,
        list_key: t.Optional[str] = None,
    ) -> None:
        models.GitHubCore.__init__(self, {}, session)
        #: Original number of items requested
        self.original: t.Final[int] = count
        #: Number of items left in the iterator
        self.count: int = count
        #: URL the class used to make it's first GET
        self.url: str = url
        #: Last URL that was requested
        self.last_url: t.Optional[str] = None
        self._api: str = self.url
        #: Class for constructing an item to return
        self.cls: t.Type[T] = cls
        #: Parameters of the query string
        self.params: t.Mapping[str, t.Optional[str]] = params or {}
        self._remove_none(self.params)
        # We do not set this from the parameter sent. We want this to
        # represent the ETag header returned by GitHub no matter what.
        # If this is not None, then it won't be set from the response and
        # that's not what we want.
        #: The ETag Header value returned by GitHub
        self.etag: t.Optional[str] = None
        #: Headers generated for the GET request
        self.headers: t.Dict[str, str] = dict(headers or {})
        #: The last response seen
        self.last_response: "requests.models.Response" = None
        #: Last status code received
        self.last_status: int = 0
        #: Key to get the list of items in case a dict is returned
        self.list_key: t.Final[t.Optional[str]] = list_key

        if etag:
            self.headers.update({"If-None-Match": etag})

        self.path: str = urlparse(self.url).path

    def _repr(self) -> str:
        return f"<GitHubIterator [{self.count}, {self.path}]>"

    def __iter__(self) -> t.Generator[T, None, None]:
        self.last_url, params = self.url, self.params
        headers = self.headers

        if 0 < self.count <= 100 and self.count != -1:
            params["per_page"] = self.count

        if "per_page" not in params and self.count == -1:
            params["per_page"] = 100

        cls = self.cls
        if issubclass(self.cls, models.GitHubCore):
            cls = functools.partial(self.cls, session=self)

        while (self.count == -1 or self.count > 0) and self.last_url:
            response = self._get(
                self.last_url, params=params, headers=headers
            )
            self.last_response = response
            self.last_status = response.status_code
            if params:
                params = None  # rel_next already has the params

            if not self.etag and response.headers.get("ETag"):
                self.etag = response.headers.get("ETag")

            json = self._get_json(response)

            if json is None:
                break

            # Some APIs return the list of items inside a dict
            if isinstance(json, dict) and self.list_key is not None:
                try:
                    json = json[self.list_key]
                except KeyError:
                    raise exceptions.UnprocessableResponseBody(
                        "GitHub's API returned a body that could not be"
                        " handled",
                        json,
                    )

            # languages returns a single dict. We want the items.
            if isinstance(json, dict):
                if issubclass(self.cls, models.GitHubCore):
                    raise exceptions.UnprocessableResponseBody(
                        "GitHub's API returned a body that could not be"
                        " handled",
                        json,
                    )
                if json.get("ETag"):
                    del json["ETag"]
                if json.get("Last-Modified"):
                    del json["Last-Modified"]
                json = json.items()

            for i in json:
                if i is None:
                    continue
                yield cls(i)
                self.count -= 1 if self.count > 0 else 0
                if self.count == 0:
                    break

            rel_next = response.links.get("next", {})
            self.last_url = rel_next.get("url", "")

    def __next__(self) -> T:
        if not hasattr(self, "__i__"):
            self.__i__ = self.__iter__()
        return next(self.__i__)

    def _get_json(self, response: "requests.models.Response"):
        return self._json(response, 200)

    def refresh(self, conditional: bool = False) -> "GitHubIterator":
        self.count = self.original
        if conditional and self.etag:
            self.headers["If-None-Match"] = self.etag
        self.etag = None
        self.__i__ = self.__iter__()
        return self

    def next(self) -> T:
        return self.__next__()


class SearchIterator(GitHubIterator):

    """This is a special-cased class for returning iterable search results.

    It inherits from :class:`GitHubIterator <github3.structs.GitHubIterator>`.
    All members and methods documented here are unique to instances of this
    class. For other members and methods, check its parent class.

    """

    _ratelimit_resource = "search"

    def __init__(
        self,
        count: int,
        url: str,
        cls: t.Type[T],
        session: "session.GitHubSession",
        params: t.Optional[t.Mapping[str, t.Optional[str]]] = None,
        etag: t.Optional[str] = None,
        headers: t.Optional[t.Mapping[str, str]] = None,
    ):
        super().__init__(count, url, cls, session, params, etag, headers)
        #: Total count returned by GitHub
        self.total_count: int = 0
        #: Items array returned in the last request
        self.items: t.List[t.Mapping[str, t.Any]] = []

    def _repr(self):
        return "<SearchIterator [{}, {}?{}]>".format(
            self.count, self.path, urlencode(self.params)
        )

    def _get_json(self, response):
        json = self._json(response, 200)
        # I'm not sure if another page will retain the total_count attribute,
        # so if it's not in the response, just set it back to what it used to
        # be
        self.total_count = json.get("total_count", self.total_count)
        self.items = json.get("items", [])
        # If we return None then it will short-circuit the while loop.
        return json.get("items")



"""Repository traffic stats logic."""
from .. import models


class ViewsStats(models.GitHubCore):
    """The total number of repository views, per day or week, for the last 14
    days.

    See also https://developer.github.com/v3/repos/traffic/

    .. note::

        Timestamps are aligned to UTC midnight of the beginning of the day or
        week. Week begins on Monday.

    This object has the following attributes:

    .. attribute:: count

        The total number of views.

    .. attribute:: uniques

        The total number of unique views.

    .. attribute:: views

        A list of dictionaries containing daily or weekly statistical data.

    """

    def _update_attributes(self, stats_object):
        self.count = stats_object["count"]
        self.uniques = stats_object["uniques"]
        self.views = stats_object["views"]
        if self.views:
            for view in self.views:
                view["timestamp"] = self._strptime(view["timestamp"])

    def _repr(self):
        return (
            "<Views Statistics " "[{s.count}, {s.uniques} unique]>"
        ).format(s=self)


class ClonesStats(models.GitHubCore):
    """The total number of repository clones, per day or week, for the last 14
    days.

    See also https://developer.github.com/v3/repos/traffic/

    .. note::

        Timestamps are aligned to UTC midnight of the beginning of the day or
        week. Week begins on Monday.

    This object has the following attributes:

    .. attribute:: count

        The total number of clones.

    .. attribute:: uniques

        The total number of unique clones.

    .. attribute:: clones

        A list of dictionaries containing daily or weekly statistical data.

    """

    def _update_attributes(self, stats_object):
        self.count = stats_object["count"]
        self.uniques = stats_object["uniques"]
        self.clones = stats_object["clones"]
        if self.clones:
            for clone in self.clones:
                clone["timestamp"] = self._strptime(clone["timestamp"])

    def _repr(self):
        return (
            "<Clones Statistics " "[{s.count}, {s.uniques} unique]>"
        ).format(s=self)



from .code import CodeSearchResult
from .commit import CommitSearchResult
from .issue import IssueSearchResult
from .repository import RepositorySearchResult
from .user import UserSearchResult


__all__ = (
    "CodeSearchResult",
    "CommitSearchResult",
    "IssueSearchResult",
    "RepositorySearchResult",
    "UserSearchResult",
)



from json import dumps

from . import label
from .. import users
from ..decorators import requires_auth
from ..models import GitHubCore


class Milestone(GitHubCore):
    """Representation of milestones on a repository.

    See also: http://developer.github.com/v3/issues/milestones/

    This object has the following attributes:

    .. attribute:: closed_issues_count

        The number of closed issues in this milestone.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        when this milestone was created.

    .. attribute:: creator

        If present, a :class:`~github3.users.ShortUser` representing the user
        who created this milestone.

    .. attribute:: description

        The written description of this milestone and its purpose.

    .. attribute:: due_on

        If set, a :class:`~datetime.datetime` object representing the date and
        time when this milestone is due.

    .. attribute:: id

        The unique identifier of this milestone in GitHub.

    .. attribute:: number

        The repository-local numeric identifier of this milestone. This starts
        at 1 like issues.

    .. attribute:: open_issues_count

        The number of open issues still in this milestone.

    .. attribute:: state

        The state of this milestone, e.g., ``'open'`` or ``'closed'``.

    .. attribute:: title

        The title of this milestone.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        when this milestone was last updated.
    """

    def _update_attributes(self, milestone):
        self._api = milestone["url"]
        self.closed_issues_count = milestone["closed_issues"]
        self.closed_issues = self.closed_issues_count
        self.created_at = self._strptime(milestone["created_at"])
        self.creator = milestone["creator"]
        if self.creator:
            self.creator = users.ShortUser(self.creator, self)
        self.description = milestone["description"]
        self.due_on = self._strptime(milestone["due_on"])
        self.id = milestone["id"]
        self.number = milestone["number"]
        self.open_issues_count = milestone["open_issues"]
        self.open_issues = self.open_issues_count
        self.state = milestone["state"]
        self.title = milestone["title"]
        self.updated_at = self._strptime(milestone["updated_at"])

    def _repr(self):
        return f"<Milestone [{self}]>"

    def __str__(self):
        return self.title

    @requires_auth
    def delete(self):
        """Delete this milestone.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    def labels(self, number=-1, etag=None):
        """Iterate over the labels of every associated issue.

        .. versionchanged:: 0.9

            Add etag parameter.

        :param int number:
            (optional), number of labels to return. Default: -1 returns all
            available labels.
        :param str etag:
            (optional), ETag header from a previous response
        :returns:
            generator of labels
        :rtype:
            :class:`~github3.issues.label.ShortLabel`
        """
        url = self._build_url("labels", base_url=self._api)
        return self._iter(int(number), url, label.ShortLabel, etag=etag)

    @requires_auth
    def update(self, title=None, state=None, description=None, due_on=None):
        """Update this milestone.

        All parameters are optional, but it makes no sense to omit all of them
        at once.

        :param str title:
            (optional), new title of the milestone
        :param str state:
            (optional), ('open', 'closed')
        :param str description:
            (optional)
        :param str due_on:
            (optional), ISO 8601 time format: YYYY-MM-DDTHH:MM:SSZ
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        data = {
            "title": title,
            "state": state,
            "description": description,
            "due_on": due_on,
        }
        self._remove_none(data)
        json = None

        if data:
            json = self._json(self._patch(self._api, data=dumps(data)), 200)
        if json:
            self._update_attributes(json)
            return True
        return False



"""This module contains the RepoTag object for GitHub's tag API."""
from . import commit
from .. import models


class RepoTag(models.GitHubCore):
    """Representation of a tag made on a GitHub repository.

    .. note::

        This is distinct from :class:`~github3.git.Tag`. This object
        has information specific to a tag on a *GitHub* repository.
        That includes URLs to the tarball and zipball files auto-generated
        by GitHub.

    See also: http://developer.github.com/v3/repos/#list-tags

    This object has the following attributes:

    .. attribute:: commit

        .. versionchanged:: 1.0.0

            This attribute used to be a two item dictionary.

        A :class:`~github3.repos.commit.MiniCommit` object representing the
        commit this tag references.

    .. attribute:: name

        The name of this tag, e.g., ``v1.0.0``.

    .. attribute:: tarball_url

        The URL for the tarball file generated by GitHub for this tag.

    .. attribute:: zipball_url

        The URL for the zipball file generated by GitHub for this tag.
    """

    def _update_attributes(self, tag):
        self.commit = commit.MiniCommit(tag["commit"], self)
        self.name = tag["name"]
        self.tarball_url = tag["tarball_url"]
        self.zipball_url = tag["zipball_url"]

    def _repr(self):
        return f"<Repository Tag [{self}]>"

    def __str__(self):
        return self.name



"""
github3.issues
==============

This module contains the classes related to issues.

See also: http://developer.github.com/v3/issues/
"""
from ..utils import timestamp_parameter
from .issue import Issue
from .issue import ShortIssue

__all__ = ["Issue", "ShortIssue"]


def issue_params(filter, state, labels, sort, direction, since):
    params = {}
    if filter in ("assigned", "created", "mentioned", "subscribed", "all"):
        params["filter"] = filter

    if state in ("open", "closed", "all"):
        params["state"] = state

    if labels:
        params["labels"] = labels

    if sort in ("created", "updated", "comments"):
        params["sort"] = sort

    if direction in ("asc", "desc"):
        params["direction"] = direction

    since = timestamp_parameter(since)
    if since:
        params["since"] = since

    return params



"""This module contains all the classes relating to pull requests."""
from json import dumps

from uritemplate import URITemplate

from . import models
from . import users
from .decorators import requires_auth
from .issues import Issue
from .issues.comment import IssueComment
from .repos import commit as rcommit
from .repos import contents


class PullDestination(models.GitHubCore):
    """The object that represents a pull request destination.

    This is the base class for the :class:`~github3.pulls.Head` and
    :class:`~github3.pulls.Base` objects. Each has identical attributes to
    this object.

    Please see GitHub's `Pull Request Documentation`_ for more information.

    .. _Pull Request Documentation:
        http://developer.github.com/v3/pulls/#get-a-single-pull-request

    .. attribute:: ref

        The full reference string for the git object

    .. attribute:: label

        The label for the destination (e.g., 'master', 'mybranch')

    .. attribute:: user

        If provided, a :class:`~github3.users.ShortUser` instance representing
        the owner of the destination

    .. attribute:: sha

        The SHA of the commit at the head of the destination

    .. attribute:: repository

        A :class:`~github3.repos.repo.ShortRepository` representing the
        repository containing this destination

    .. attribute:: repo

        A tuple containing the login and repository name, e.g.,
        ('sigmavirus24', 'github3.py')

        This attribute is generated by github3.py and may be deprecated in the
        future.
    """

    direction = None

    def _update_attributes(self, dest):
        from .repos.repo import ShortRepository

        self.ref = dest["ref"]
        self.label = dest["label"]
        self.user = dest.get("user")
        if self.user:
            self.user = users.ShortUser(self.user, self)
        self.sha = dest["sha"]
        self._repo_name = ""
        self._repo_owner = ""
        repo = dest.get("repo")
        if repo:
            self._repo_name = repo.get("name")
            self._repo_owner = repo["owner"]["login"]
            self.repository = ShortRepository(repo, self)
        self.repo = (self._repo_owner, self._repo_name)

    def _repr(self):
        return f"<{self.direction} [{self.label}]>"


class Head(PullDestination):
    """An object representing the Head destination of a pull request.

    See :class:`~github3.pulls.PullDestination` for more details.
    """

    destination = "Head"


class Base(PullDestination):
    """An object representing the Base destination of a pull request.

    See :class:`~github3.pulls.PullDestination` for more details.
    """

    destination = "Base"


class PullFile(models.GitHubCore):
    """The object that represents a file in a pull request.

    Please see GitHub's `Pull Request Files Documentation`_ for more
    information.

    .. _Pull Request Files Documentation:
        http://developer.github.com/v3/pulls/#list-pull-requests-files

    .. attribute:: additions_count

        The number of additions made to this file

    .. attribute:: blob_url

        The API resource used to retrieve the blob for this file

    .. attribute:: changes_count

        The number of changes made to this file

    .. attribute:: contents_url

        The API resource to view the raw contents of this file

    .. attribute:: deletions_count

        The number of deletions made to this file

    .. attribute:: filename

        The name of this file

    .. attribute:: patch

        The patch generated by this

        .. note::

            If the patch is larger than a specific size it may be missing
            from GitHub's response. The attribute will be set to ``None``
            in this case.

    .. attribute:: raw_url

        The API resource to view the raw diff of this file

    .. attribute:: sha

        The SHA of the commit that this file belongs to

    .. attribute:: status

        The string with the status of the file, e.g., 'added'
    """

    def _update_attributes(self, pfile):
        self.sha = pfile["sha"]
        self.filename = pfile["filename"]
        self.status = pfile["status"]
        self.additions_count = pfile["additions"]
        self.deletions_count = pfile["deletions"]
        self.changes_count = pfile["changes"]
        self.blob_url = pfile["blob_url"]
        self.raw_url = pfile["raw_url"]
        self.patch = pfile.get("patch")
        self.contents_url = pfile["contents_url"]

    def _repr(self):
        return f"<Pull Request File [{self.filename}]>"

    def contents(self):
        """Return the contents of the file.

        :returns:
            An object representing the contents of this file
        :rtype:
            :class:`~github3.repos.contents.Contents`
        """
        json = self._json(self._get(self.contents_url), 200)
        return self._instance_or_null(contents.Contents, json)


class _PullRequest(models.GitHubCore):
    """The :class:`PullRequest <PullRequest>` object.

    Please see GitHub's `PullRequests Documentation`_ for more information.

    .. _PullRequests Documentation:
        http://developer.github.com/v3/pulls/
    """

    class_name = "_PullRequest"

    def _update_attributes(self, pull):
        from . import orgs

        self._api = pull["url"]
        self.active_lock_reason = pull["active_lock_reason"]
        self.assignee = pull["assignee"]
        if self.assignee is not None:
            self.assignee = users.ShortUser(self.assignee, self)
        self.assignees = [users.ShortUser(a, self) for a in pull["assignees"]]
        self.base = Base(pull["base"], self)
        self.body = pull["body"]
        self.body_html = pull["body_html"]
        self.body_text = pull["body_text"]
        self.closed_at = self._strptime(pull["closed_at"])
        self.comments_url = pull["comments_url"]
        self.commits_url = pull["commits_url"]
        self.created_at = self._strptime(pull["created_at"])
        self.diff_url = pull["diff_url"]
        self.head = Head(pull["head"], self)
        self.html_url = pull["html_url"]
        self.id = pull["id"]
        self.issue_url = pull["issue_url"]
        self.links = pull["_links"]
        self.locked = pull["locked"]
        self.merge_commit_sha = pull["merge_commit_sha"]
        self.merged_at = self._strptime(pull["merged_at"])
        self.number = pull["number"]
        self.patch_url = pull["patch_url"]
        self.rebaseable = pull.get("rebaseable")
        requested_reviewers = pull.get("requested_reviewers", [])
        self.requested_reviewers = [
            users.ShortUser(r, self) for r in requested_reviewers
        ]
        requested_teams = pull.get("requested_teams", [])
        self.requested_teams = [
            orgs.ShortTeam(t, self) for t in requested_teams
        ]
        self.review_comment_urlt = URITemplate(pull["review_comment_url"])
        self.review_comments_url = pull["review_comments_url"]
        self.repository = None
        if self.base:
            self.repository = self.base.repository
        self.state = pull["state"]
        self.statuses_url = pull["statuses_url"]
        self.title = pull["title"]
        self.updated_at = self._strptime(pull["updated_at"])
        self.user = users.ShortUser(pull["user"], self)

    def _repr(self):
        return f"<{self.class_name} [#{self.number}]>"

    @requires_auth
    def close(self):
        """Close this Pull Request without merging.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self.update(self.title, self.body, "closed")

    @requires_auth
    def create_comment(self, body):
        """Create a comment on this pull request's issue.

        :param str body:
            (required), comment body
        :returns:
            the comment that was created on the pull request
        :rtype:
            :class:`~github3.issues.comment.IssueComment`
        """
        url = self.comments_url
        json = None
        if body:
            json = self._json(self._post(url, data={"body": body}), 201)
        return self._instance_or_null(IssueComment, json)

    @requires_auth
    def create_review_comment(self, body, commit_id, path, position):
        """Create a review comment on this pull request.

        .. note::

            All parameters are required by the GitHub API.

        :param str body:
            The comment text itself
        :param str commit_id:
            The SHA of the commit to comment on
        :param str path:
            The relative path of the file to comment on
        :param int position:
            The line index in the diff to comment on.
        :returns:
            The created review comment.
        :rtype:
            :class:`~github3.pulls.ReviewComment`
        """
        url = self._build_url("comments", base_url=self._api)
        data = {
            "body": body,
            "commit_id": commit_id,
            "path": path,
            "position": int(position),
        }
        json = self._json(self._post(url, data=data), 201)
        return self._instance_or_null(ReviewComment, json)

    @requires_auth
    def create_review_requests(self, reviewers=None, team_reviewers=None):
        """Ask for reviews on this pull request.

        :param list reviewers:
            The users to which request a review
        :param list team_reviewers:
            The teams to which request a review
        :returns:
            The pull request on which the reviews were requested
        :rtype:
            :class:`~github3.pulls.ShortPullRequest`
        """
        url = self._build_url("requested_reviewers", base_url=self._api)
        data = {}
        if reviewers is not None:
            data["reviewers"] = [getattr(r, "login", r) for r in reviewers]
        if team_reviewers is not None:
            data["team_reviewers"] = [
                getattr(t, "slug", t) for t in team_reviewers
            ]
        json = self._json(self._post(url, data=data), 201)
        return self._instance_or_null(ShortPullRequest, json)

    @requires_auth
    def create_review(self, body, commit_id=None, event=None, comments=None):
        """Create a review comment on this pull request.

        .. warning::

            If you do not specify ``event``, GitHub will default it
            to ``PENDING`` which means that your review will need to
            be submitted after creation. (See also
            :meth:`~github3.pulls.PullReview.submit`.)

        :param str body:
            The comment text itself, required when using COMMENT or
            REQUEST_CHANGES.
        :param str commit_id:
            The SHA of the commit to comment on
        :param str event:
            The review action you want to perform. Actions include
            APPROVE, REQUEST_CHANGES or COMMENT. By leaving this blank
            you set the action to PENDING and will need to submit the
            review. Leaving blank may result in a 422 error response
            which will need to be handled.
        :param list comments:
            Array of draft review comment objects. Please see Github's
            `Create a pull request review documentation`_ for details
            on review comment objects. At the time of writing these
            were a dictionary with 3 keys: `path`, `position` and
            `body`.

            .. _Create a pull request review documentation:
                https://developer.github.com/v3/pulls/reviews/#create-a-pull-request-review

        :returns:
            The created review.
        :rtype:
            :class:`~github3.pulls.PullReview`
        """
        if comments is None:
            comments = []
        url = self._build_url("reviews", base_url=self._api)
        data = {"body": body, "comments": comments}
        if commit_id is not None:
            data["commit_id"] = commit_id
        if event is not None:
            data["event"] = event
        json = self._json(self._post(url, data=data), 200)
        return self._instance_or_null(PullReview, json)

    @requires_auth
    def delete_review_requests(self, reviewers=None, team_reviewers=None):
        """Cancel review requests on this pull request.

        :param list reviewers:
            The users whose review is no longer requested
        :param list team_reviewers:
            The teams whose review is no longer requested
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("requested_reviewers", base_url=self._api)
        data = {}
        if reviewers is not None:
            data["reviewers"] = [getattr(r, "login", r) for r in reviewers]
        if team_reviewers is not None:
            data["team_reviewers"] = [
                getattr(t, "slug", t) for t in team_reviewers
            ]
        return self._boolean(self._delete(url, data=dumps(data)), 200, 404)

    def diff(self):
        """Return the diff.

        :returns:
            representation of the diff
        :rtype:
            bytes
        """
        resp = self._get(
            self._api, headers={"Accept": "application/vnd.github.diff"}
        )
        return resp.content if self._boolean(resp, 200, 404) else b""

    def is_merged(self):
        """Check to see if the pull request was merged.

        .. versionchanged:: 1.0.0

            This now always makes a call to the GitHub API. To avoid that,
            check :attr:`merged` before making this call.

        :returns:
            True if merged, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("merge", base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    def issue(self):
        """Retrieve the issue associated with this pull request.

        :returns:
            the issue object that this pull request builds upon
        :rtype:
            :class:`~github3.issues.Issue`
        """
        json = self._json(self._get(self.issue_url), 200)
        return self._instance_or_null(Issue, json)

    def commits(self, number=-1, etag=None):
        """Iterate over the commits on this pull request.

        :param int number:
            (optional), number of commits to return. Default: -1 returns all
            available commits.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of repository commit objects
        :rtype:
            :class:`~github3.repos.commit.ShortCommit`
        """
        url = self._build_url("commits", base_url=self._api)
        return self._iter(int(number), url, rcommit.ShortCommit, etag=etag)

    def files(self, number=-1, etag=None):
        """Iterate over the files associated with this pull request.

        :param int number:
            (optional), number of files to return. Default: -1 returns all
            available files.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns: generator of pull request files
        :rtype: :class:`~PullFile`
        """
        url = self._build_url("files", base_url=self._api)
        return self._iter(int(number), url, PullFile, etag=etag)

    def issue_comments(self, number=-1, etag=None):
        """Iterate over the issue comments on this pull request.

        In this case, GitHub leaks implementation details. Pull Requests are
        really just Issues with a diff. As such, comments on a pull request
        that are not in-line with code, are technically issue comments.

        :param int number:
            (optional), number of comments to return. Default: -1 returns all
            available comments.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of non-review comments on this pull request
        :rtype:
            :class:`~github3.issues.comment.IssueComment`
        """
        comments = self.links.get("comments", {})
        url = comments.get("href")
        return self._iter(int(number), url, IssueComment, etag=etag)

    @requires_auth
    def merge(
        self,
        commit_message=None,
        sha=None,
        merge_method="merge",
        commit_title=None,
    ):
        """Merge this pull request.

        .. versionchanged:: 1.3.0

            The ``commit_title`` parameter has been added to allow users to
            set the merge commit title.

        .. versionchanged:: 1.0.0

            The boolean ``squash`` parameter has been replaced with
            ``merge_method`` which requires a string.

        :param str commit_message:
            (optional), message to be used for the merge commit
        :param str commit_title:
            (optional), message to be used for the merge commit title
        :param str sha:
            (optional), SHA that pull request head must match to merge.
        :param str merge_method: (optional), Change the merge method.
            Either 'merge', 'squash' or 'rebase'. Default is 'merge'.
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        :returns: bool
        """
        parameters = {"merge_method": merge_method}
        if sha:
            parameters["sha"] = sha
        if commit_message is not None:
            parameters["commit_message"] = commit_message
        if commit_title is not None:
            parameters["commit_title"] = commit_title
        url = self._build_url("merge", base_url=self._api)
        json = self._json(self._put(url, data=dumps(parameters)), 200)
        if not json:
            return False
        return json["merged"]

    def patch(self):
        """Return the patch.

        :returns:
            bytestring representation of the patch
        :rtype:
            bytes
        """
        resp = self._get(
            self._api, headers={"Accept": "application/vnd.github.patch"}
        )
        return resp.content if self._boolean(resp, 200, 404) else b""

    @requires_auth
    def reopen(self):
        """Re-open a closed Pull Request.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self.update(self.title, self.body, state="open")

    def review_comments(self, number=-1, etag=None):
        """Iterate over the review comments on this pull request.

        :param int number:
            (optional), number of comments to return. Default: -1 returns all
            available comments.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of review comments
        :rtype:
            :class:`~github3.pulls.ReviewComment`
        """
        url = self._build_url("comments", base_url=self._api)
        return self._iter(int(number), url, ReviewComment, etag=etag)

    def review_requests(self):
        """Retrieve the review requests associated with this pull request.

        :returns:
            review requests associated with this pull request
        :rtype:
            :class:`~github3.pulls.ReviewRequests`
        """
        url = self._build_url("requested_reviewers", base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(ReviewRequests, json)

    def reviews(self, number=-1, etag=None):
        """Iterate over the reviews associated with this pull request.

        :param int number:
            (optional), number of reviews to return. Default: -1 returns all
            available files.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of reviews for this pull request
        :rtype:
            :class:`~github3.pulls.PullReview`
        """
        url = self._build_url("reviews", base_url=self._api)
        return self._iter(int(number), url, PullReview, etag=etag)

    @requires_auth
    def update(
        self,
        title=None,
        body=None,
        state=None,
        base=None,
        maintainer_can_modify=None,
    ):
        """Update this pull request.

        :param str title:
            (optional), title of the pull
        :param str body:
            (optional), body of the pull request
        :param str state:
            (optional), one of ('open', 'closed')
        :param str base:
            (optional), Name of the branch on the current repository that the
            changes should be pulled into.
        :param bool maintainer_can_modify:
            (optional), Indicates whether a maintainer is allowed to modify the
            pull request or not.
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        data = {
            "title": title,
            "body": body,
            "state": state,
            "base": base,
            "maintainer_can_modify": maintainer_can_modify,
        }
        json = None
        self._remove_none(data)

        if data:
            json = self._json(self._patch(self._api, data=dumps(data)), 200)

        if json:
            self._update_attributes(json)
            return True
        return False


class PullRequest(_PullRequest):
    """Object for the full representation of a PullRequest.

    GitHub's API returns different amounts of information about prs based
    upon how that information is retrieved. This object exists to represent
    the full amount of information returned for a specific pr. For example,
    you would receive this class when calling
    :meth:`~github3.github.GitHub.pull_request`. To provide a clear
    distinction between the types of prs, github3.py uses different classes
    with different sets of attributes.

    .. versionchanged:: 1.0.0

    This object has all of the same attributes as
    :class:`~github3.pulls.ShortPullRequest` as well as the following:

    .. attribute:: additions_count

        The number of lines of code added in this pull request.

    .. attribute:: deletions_count

        The number of lines of code deleted in this pull request.

    .. attribute:: comments_count

        The number of comments left on this pull request.

    .. attribute:: commits_count

        The number of commits included in this pull request.

    .. attribute:: mergeable

        A boolean attribute indicating whether GitHub deems this pull request
        is mergeable.

    .. attribute:: mergeable_state

        A string indicating whether this would be a 'clean' or 'dirty' merge.

    .. attribute:: merged

        A boolean attribute indicating whether the pull request has been merged
        or not.

    .. attribute:: merged_by

        An instance of :class:`~github3.users.ShortUser` to indicate the user
        who merged this pull request. If this hasn't been merged or if
        :attr:`mergeable` is still being decided by GitHub this will be
        ``None``.

    .. attribute:: review_comments_count

        The number of review comments on this pull request.
    """

    class_name = "Pull Request"

    def _update_attributes(self, pull):
        super()._update_attributes(pull)
        self.additions_count = pull["additions"]
        self.auto_merge = pull.get("auto_merge", None)
        self.author_association = pull["author_association"]
        self.comments_count = pull["comments"]
        self.commits_count = pull["commits"]
        self.deletions_count = pull["deletions"]
        self.draft = pull["draft"]
        self.mergeable = pull["mergeable"]
        self.mergeable_state = pull["mergeable_state"]
        self.merged = pull["merged"]
        self.merged_by = pull["merged_by"]
        if self.merged_by is not None:
            self.merged_by = users.ShortUser(self.merged_by, self)
        self.review_comments_count = pull["review_comments"]

        requested_teams = pull["requested_teams"]
        self.requested_teams = []
        if requested_teams:
            from . import orgs

            self.requested_teams = [
                orgs.ShortTeam(team, self) for team in requested_teams
            ]

        requested_reviewers = pull["requested_reviewers"]
        self.requested_reviewers = []
        if requested_reviewers:
            self.requested_reviewers = [
                users.ShortUser(user, self) for user in requested_reviewers
            ]


class ShortPullRequest(_PullRequest):
    """Object for the shortened representation of a PullRequest.

    GitHub's API returns different amounts of information about prs based
    upon how that information is retrieved. Often times, when iterating over
    several prs, GitHub will return less information. To provide a clear
    distinction between the types of Pull Requests, github3.py uses different
    classes with different sets of attributes.

    .. versionadded:: 1.0.0

    The attributes available on this object are:

    .. attribute:: url

        The URL that describes this exact pull request.

    .. attribute:: assignee

        .. deprecated:: 1.0.0

            Use :attr:`assignees` instead.

        The assignee of the pull request, if present, represented as an
        instance of :class:`~github3.users.ShortUser`

    .. attribute:: assignees

        .. versionadded:: 1.0.0

        A list of the assignees of the pull request. If not empty, a list
        of instances of :class:`~github3.users.ShortUser`.

    .. attribute:: base

        A :class:`~github3.pulls.Base` object representing the base pull
        request destination.

    .. attribute:: body

        The Markdown formatted body of the pull request message.

    .. attribute:: body_html

        The HTML formatted body of the pull request mesage.

    .. attribute:: body_text

        The plain-text formatted body of the pull request message.

    .. attribute:: closed_at

        A :class:`~datetime.datetime` object representing the date and time
        when this pull request was closed.

    .. attribute:: comments_url

        The URL to retrieve the comments on this pull request from the API.

    .. attribute:: commits_url

        The URL to retrieve the commits in this pull request from the API.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        when this pull request was created.

    .. attribute:: diff_url

        The URL to retrieve the diff for this pull request via the API.

    .. attribute:: head

        A :class:`~github3.pulls.Head` object representing the head pull
        request destination.

    .. attribute:: html_url

        The URL one would use to view this pull request in the browser.

    .. attribute:: id

        The unique ID of this pull request across all of GitHub.

    .. attribute:: issue_url

        The URL of the resource that represents this pull request as an issue.

    .. attribute:: links

        A dictionary provided by ``_links`` in the API response.

        .. versionadded:: 1.0.0

    .. attribute:: merge_commit_sha

        If unmerged, holds the sha of the commit to test mergability.
        If merged, holds commit sha of the merge commit, squashed commit on
        the base branch or the commit that the base branch was updated to
        after rebasing the PR.

    .. attribute:: merged_at

        A :class:`~datetime.datetime` object representing the date and time
        this pull request was merged. If this pull request has not been merged
        then this attribute will be ``None``.

    .. attribute:: number

        The number of the pull request on the repository.

    .. attribute:: patch_url

        The URL to retrieve the patch for this pull request via the API.

    .. attribute:: rebaseable

        A boolean attribute indicating whether GitHub deems this pull request
        is rebaseable. None if not set.

    .. attribute:: repository

        A :class:`~github3.repos.repo.ShortRepository` from the :attr:`base`
        instance.

    .. attribute:: requested_reviewers

        A list of :class:`~github3.users.ShortUser` from which a review was
        requested.

    .. attribute:: requested_teams

       A list of :class:`~github3.orgs.ShortTeam` from which a review was
       requested.

    .. attribute:: review_comment_urlt

        A URITemplate instance that expands to provide the review comment URL
        provided a number.

    .. attribute:: review_comments_url

        The URl to retrieve all review comments on this pull request from the
        API.

    .. attribute:: state

        The current state of this pull request.

    .. attribute:: title

        The title of this pull request.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` instance representing the date and time
        when this pull request was last updated.

    .. attribute:: user

        A :class:`~github3.users.ShortUser` instance representing who opened
        this pull request.
    """

    class_name = "Short Pull Request"
    _refresh_to = PullRequest


class PullReview(models.GitHubCore):
    """Object representing a Pull Request Review returned by the API.

    Please see GitHub's `Pull Review Documentation`_ for more information.

    .. _PullReview Documentation:
        https://developer.github.com/v3/pulls/reviews/

    .. attribute:: id

        The unique ID of this pull request review.

    .. attribute:: author_association

        .. versionadded:: 1.0.0

        The relationship of this review's author to the project.

    .. attribute:: body

        The Markdown formatted body of this review.

    .. attribute:: body_html

        .. versionadded:: 1.0.0

        The HTML formatted body of this review.

    .. attribute:: body_text

        .. versionadded:: 1.0.0

        The plain-text formatted body of this review.

    .. attribute:: commit_id

        The SHA of the commit that the review was left on.

        .. note::

            It is possible for the attribute to be set to ``None``, if the
            review references a commit that is no longer available in the pull
            request branch, such as after a force push.

    .. attribute:: html_url

        .. versionadded:: 1.0.0

        The URL to view this pull request in a browser.

    .. attribute:: pull_request_url

        The URL to retrieve the pull request via the API.

    .. attribute:: state

        The state of this review, e.g., the option specified in the review
        dialog by the author.

    .. attribute:: submitted_at

        A :class:`~datetime.datetime` object representing the date and time
        this review was submitted.

    .. attribute:: user

        A :class:`~github3.users.ShortUser` instance representing the author
        of this review.
    """

    def _update_attributes(self, review):
        self.id = review["id"]
        self.author_association = review["author_association"]
        self.body = review["body"]
        self.body_html = review["body_html"]
        self.body_text = review["body_text"]
        # NOTE(pabelanger): In some cases, commit_id could be missing on a
        # PullReview.
        self.commit_id = review.get("commit_id", None)
        self.html_url = review["html_url"]
        self.user = users.ShortUser(review["user"], self)
        self.state = review["state"]
        self.submitted_at = self._strptime(review.get("submitted_at"))
        self.pull_request_url = review["pull_request_url"]

    def _repr(self):
        return f"<Pull Request Review [{self.id}]>"

    @requires_auth
    def submit(self, body, event=None):
        """Submit a pull request review.

        :param str body:
            The body text of the pull request review.
        :param str event:
            The review action you want to perform. Actions include
            APPROVE, REQUEST_CHANGES or COMMENT. By leaving this blank
            you set the action to PENDING and will need to submit the
            review. Leaving blank will result in a 422 error response
            which will need to be handled.
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url(
            "reviews", self.id, "events", base_url=self.pull_request_url
        )
        data = {"body": body}
        if event is not None:
            data["event"] = event
        json = self._json(self._post(url, data=data), 200)
        if json:
            self._update_attributes(json)
            return True
        return False


class ReviewComment(models.GitHubCore):
    """Object representing review comments left on a pull request.

    Please see GitHub's `Pull Comments Documentation`_ for more information.

    .. _Pull Comments Documentation:
        http://developer.github.com/v3/pulls/comments/

    .. attribute:: id

        The unique identifier for this comment across all GitHub review
        comments.

    .. attribute:: author_association

        The role of the author of this comment on the repository.

    .. attribute:: body

        The Markdown formatted body of this comment.

    .. attribute:: body_html

        The HTML formatted body of this comment.

    .. attribute:: body_text

        The plain text formatted body of this comment.

    .. attribute:: commit_id

        The SHA of current commit this comment was left on.

    .. attribute:: created_at

        A :class:`~datetime.datetime` instance representing the date and time
        this comment was created.

    .. attribute:: diff_hunk

        A string representation of the hunk of the diff where the comment was
        left.

    .. attribute:: html_url

        The URL to view this comment in the webbrowser.

    .. attribute:: links

        A dictionary of relevant URLs usually returned in the ``_links``
        attribute.

    .. attribute:: original_commit_id

        The SHA of the original commit this comment was left on.

    .. attribute:: original_position

        The original position within the diff that this comment was left on.

    .. attribute:: pull_request_url

        The URL to retrieve the pull request via the API.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` instance representing the date and time
        this comment was updated.

    .. attribute:: user

        A :class:`~github3.users.ShortUser` instance representing the author
        of this comment.

    """

    def _update_attributes(self, comment):
        self._api = comment["url"]
        self.id = comment["id"]
        self.author_association = comment["author_association"]
        self.body = comment["body"]
        self.body_html = comment["body_html"]
        self.body_text = comment["body_text"]
        self.commit_id = comment["commit_id"]
        self.created_at = self._strptime(comment["created_at"])
        self.diff_hunk = comment["diff_hunk"]
        self.html_url = comment["html_url"]
        self.links = comment["_links"]
        self.original_commit_id = comment["original_commit_id"]
        self.original_position = comment["original_position"]
        self.path = comment["path"]
        self.position = comment["position"]
        self.pull_request_url = comment["pull_request_url"]
        self.updated_at = self._strptime(comment["updated_at"])
        self.user = users.ShortUser(comment["user"], self)

    def _repr(self):
        return f"<Review Comment [{self.user.login}]>"

    @requires_auth
    def delete(self):
        """Delete this comment.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def edit(self, body):
        """Edit this comment.

        :param str body:
            (required), new body of the comment, Markdown formatted
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        if body:
            json = self._json(
                self._patch(self._api, data=dumps({"body": body})), 200
            )
            if json:
                self._update_attributes(json)
                return True
        return False

    @requires_auth
    def reply(self, body):
        """Reply to this review comment with a new review comment.

        :param str body:
            The text of the comment.
        :returns:
            The created review comment.
        :rtype:
            :class:`~github3.pulls.ReviewComment`
        """
        url = self._build_url("comments", base_url=self.pull_request_url)
        index = self._api.rfind("/") + 1
        in_reply_to = int(self._api[index:])
        json = self._json(
            self._post(url, data={"body": body, "in_reply_to": in_reply_to}),
            201,
        )
        return self._instance_or_null(ReviewComment, json)


class ReviewRequests(models.GitHubCore):
    """Object representing review requests in the GitHub API.

    .. attribute:: teams

        The list of teams that were requested a review

    .. attribute:: users

        The list of users that were requested a review

    Please see GitHub's `Review Request Documentation`_ for more information.

    .. _Review Request Documentation:
       https://developer.github.com/v3/pulls/review_requests/
    """

    def _update_attributes(self, requests):
        from . import orgs

        self.teams = [orgs.ShortTeam(t, self) for t in requests["teams"]]
        self.users = [users.ShortUser(u, self) for u in requests["users"]]

    def _repr(self):
        return "<Review Requests [users: {}, teams: {}]>".format(
            len(self.users), len(self.teams)
        )



"""This module contains everything relating to Users."""
import typing as t
from json import dumps

from uritemplate import URITemplate

from . import models
from .decorators import requires_auth
from .events import Event
from github3.auths import Authorization


class GPGKey(models.GitHubCore):
    """The object representing a user's GPG key.

    .. versionadded:: 1.2.0

    Please see GitHub's `GPG Key Documentation` for more information.

    .. _GPG Key Documentation:
        https://developer.github.com/v3/users/gpg_keys/

    .. attribute:: can_certify

        Whether this GPG key can be used to sign a key.

    .. attribute:: can_encrypt_comms

        Whether this GPG key can be used to encrypt communications.

    .. attribute:: can_encrypt_storage

        Whether this GPG key can be used to encrypt storage.

    .. attribute:: can_sign

        Whether this GPG key can be used to sign some data.

    .. attribute:: created_at

        A :class:`~datetime.datetime` representing the date and time when
        this GPG key was created.

    .. attribute:: emails

        A list of :class:`~github3.users.ShortEmail` attached to this GPG
        key.

    .. attribute:: expires_at

        A :class:`~datetime.datetime` representing the date and time when
        this GPG key will expire.

    .. attribute:: id

        The unique identifier of this GPG key.

    .. attribute:: key_id

        A hexadecimal string that identifies this GPG key.

    .. attribute:: primary_key_id

        The unique identifier of the primary key of this GPG key.

    .. attribute:: public_key

        The public key contained in this GPG key. This is not a GPG formatted
        key, and is not suitable to be used directly in programs like GPG.

    .. attribute:: subkeys

        A list of :class:`~github3.users.GPGKey` of the subkeys of this GPG
        key.
    """

    def _update_attributes(self, key):
        self.can_certify = key["can_certify"]
        self.can_encrypt_comms = key["can_encrypt_comms"]
        self.can_encrypt_storage = key["can_encrypt_storage"]
        self.can_sign = key["can_sign"]
        self.created_at = self._strptime(key["created_at"])
        self.emails = [ShortEmail(email, self) for email in key["emails"]]
        self.expires_at = self._strptime(key["expires_at"])
        self.id = key["id"]
        self.key_id = key["key_id"]
        self.primary_key_id = key["primary_key_id"]
        self.public_key = key["public_key"]
        self.subkeys = [GPGKey(subkey, self) for subkey in key["subkeys"]]

    def _repr(self):
        return f"<GPG Key [{self.key_id}]>"

    def __str__(self):
        return self.key_id

    @requires_auth
    def delete(self):
        """Delete this GPG key.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("user", "gpg_keys", self.id)
        return self._boolean(self._delete(url), 204, 404)


class Key(models.GitHubCore):
    """The object representing a user's SSH key.

    Please see GitHub's `Key Documentation`_ for more information.

    .. _Key Documentation:
        http://developer.github.com/v3/users/keys/

    .. versionchanged:: 1.0.0

        Removed ``title`` attribute

    .. attribute:: key

        A string containing the actual text of the SSH Key

    .. attribute:: id

        GitHub's unique ID for this key
    """

    def _update_attributes(self, key, session=None):
        self._api = key.get("url")
        self.key = key["key"]
        self.id = key["id"]

    def _repr(self):
        return f"<User Key [{self.title}]>"

    def __str__(self):
        return self.key

    @requires_auth
    def delete(self):
        """Delete this key."""
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def update(self, title, key):
        """Update this key.

        .. warning::

            As of 20 June 2014, the API considers keys to be immutable.
            This will soon begin to return MethodNotAllowed errors.

        :param str title: (required), title of the key
        :param str key: (required), text of the key file
        :returns: bool
        """
        json = None
        if title and key:
            data = {"title": title, "key": key}
            json = self._json(self._patch(self._api, data=dumps(data)), 200)
        if json:
            self._update_attributes(json)
            return True
        return False


class Plan(models.GitHubCore):
    """The :class:`Plan <Plan>` object.

    Please see GitHub's `Authenticated User`_ documentation for more details.

    .. _Authenticated User:
        http://developer.github.com/v3/users/#get-the-authenticated-user

    .. attribute:: collaborators_count

        .. versionchanged:: 1.0.0

        The number of collaborators allowed on this plan

    .. attribute:: name

        The name of the plan on GitHub

    .. attribute:: private_repos_count

        .. versionchanged:: 1.0.0

        The number of allowed private repositories

    .. attribute:: space

        The amount of space allotted by this plan
    """

    def _update_attributes(self, plan):
        self.collaborators = plan["collaborators"]
        self.name = plan["name"]
        self.private_repos_count = plan["private_repos"]
        self.space = plan["space"]

    def _repr(self):
        return f"<Plan [{self.name}]>"  # (No coverage)

    def __str__(self):
        return self.name

    def is_free(self):
        """Check if this is a free plan.

        :returns: bool
        """
        return self.name == "free"  # (No coverage)


class _Email(models.GitHubCore):
    """Base email object."""

    class_name = "_Email"

    def _update_attributes(self, email):
        self.email = email["email"]
        self.verified = email["verified"]

    def _repr(self):
        return f"<{self.class_name} [{self.email}]>"

    def __str__(self):
        return self.email


class ShortEmail(_Email):
    """The object used to represent an email attached to a GPG key.

    This object has the following attributes:

    .. attribute:: email

        The email address as a string

    .. attribute:: verified

        A boolean value representing whether the address has been verified or
        not
    """

    class_name = "ShortEmail"


class Email(_Email):
    """The object used to represent an AuthenticatedUser's email.

    Please see GitHub's `Emails documentation`_ for more information.

    .. _Emails documentation:
        https://developer.github.com/v3/users/emails/

    This object has all of the attributes of :class:`ShortEmail` as well as
    the following attributes:

    .. attribute:: primary

        A boolean value representing whether the address is the primary
        address for the user or not

    .. attribute:: visibility

        A string value representing whether an authenticated user can view the
        email address. Use ``public`` to allow it, ``private`` to disallow it.
    """

    def _update_attributes(self, email):
        super()._update_attributes(email)
        self.primary = email["primary"]
        self.visibility = email["visibility"]

    def _repr(self):
        return f"<Email [{self.email}]>"

    def __str__(self):
        return self.email


class _User(models.GitHubCore):
    """The :class:`User <User>` object.

    This handles and structures information in the `User section`_.

    Two user instances can be checked like so::

        u1 == u2
        u1 != u2

    And is equivalent to::

        u1.id == u2.id
        u1.id != u2.id

    .. _User section:
        http://developer.github.com/v3/users/
    """

    class_name = "_User"

    def _update_attributes(self, user):
        self.avatar_url = user["avatar_url"]
        self.events_urlt = URITemplate(user["events_url"])
        self.followers_url = user["followers_url"]
        self.following_urlt = URITemplate(user["following_url"])
        self.gists_urlt = URITemplate(user["gists_url"])
        self.gravatar_id = user["gravatar_id"]
        self.html_url = user["html_url"]
        self.id = user["id"]
        self.login = user["login"]
        self.organizations_url = user["organizations_url"]
        self.received_events_url = user["received_events_url"]
        self.repos_url = user["repos_url"]
        self.site_admin = user.get("site_admin")
        self.starred_urlt = URITemplate(user["starred_url"])
        self.subscriptions_url = user["subscriptions_url"]
        self.type = user["type"]
        self.url = self._api = user["url"]
        self._uniq = self.id

    def __str__(self):
        return self.login

    def _repr(self):
        full_name = ""
        name = getattr(self, "name", None)
        if name is not None:
            full_name = f":{name}"
        return "<{s.class_name} [{s.login}{full_name}]>".format(
            s=self, full_name=full_name
        )

    def is_assignee_on(self, username, repository):
        """Check if this user can be assigned to issues on username/repository.

        :param str username: owner's username of the repository
        :param str repository: name of the repository
        :returns: True if the use can be assigned, False otherwise
        :rtype: :class:`bool`
        """
        url = self._build_url(
            "repos", username, repository, "assignees", self.login
        )
        return self._boolean(self._get(url), 204, 404)

    def is_following(self, username):
        """Check if this user is following ``username``.

        :param str username: (required)
        :returns: bool

        """
        url = self.following_urlt.expand(other_user=username)
        return self._boolean(self._get(url), 204, 404)

    def events(self, public=False, number=-1, etag=None):
        r"""Iterate over events performed by this user.

        :param bool public: (optional), only list public events for the
            authenticated user
        :param int number: (optional), number of events to return. Default: -1
            returns all available events.
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Event <github3.events.Event>`\ s
        """
        path = ["events"]
        if public:
            path.append("public")
        url = self._build_url(*path, base_url=self._api)
        return self._iter(int(number), url, Event, etag=etag)

    def followers(self, number=-1, etag=None):
        r"""Iterate over the followers of this user.

        :param int number: (optional), number of followers to return. Default:
            -1 returns all available
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`User <User>`\ s
        """
        url = self._build_url("followers", base_url=self._api)
        return self._iter(int(number), url, ShortUser, etag=etag)

    def following(self, number=-1, etag=None):
        r"""Iterate over the users being followed by this user.

        :param int number: (optional), number of users to return. Default: -1
            returns all available users
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`User <User>`\ s
        """
        url = self._build_url("following", base_url=self._api)
        return self._iter(int(number), url, ShortUser, etag=etag)

    def gpg_keys(self, number=-1, etag=None):
        r"""Iterate over the GPG keys of this user.

        .. versionadded:: 1.2.0

        :param int number: (optional), number of GPG keys to return. Default:
            -1 returns all available GPG keys
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`GPGKey <GPGKey>`\ s
        """
        url = self._build_url("gpg_keys", base_url=self._api)
        return self._iter(int(number), url, GPGKey, etag=etag)

    def keys(self, number=-1, etag=None):
        r"""Iterate over the public keys of this user.

        .. versionadded:: 0.5

        :param int number: (optional), number of keys to return. Default: -1
            returns all available keys
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Key <Key>`\ s
        """
        url = self._build_url("keys", base_url=self._api)
        return self._iter(int(number), url, Key, etag=etag)

    @requires_auth
    def organization_events(self, org, number=-1, etag=None):
        r"""Iterate over events from the user's organization dashboard.

        .. note:: You must be authenticated to view this.

        :param str org: (required), name of the organization
        :param int number: (optional), number of events to return. Default: -1
            returns all available events
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Event <github3.events.Event>`\ s
        """
        url = ""
        if org:
            url = self._build_url("events", "orgs", org, base_url=self._api)
        return self._iter(int(number), url, Event, etag=etag)

    def received_events(self, public=False, number=-1, etag=None):
        r"""Iterate over events that the user has received.

        If the user is the authenticated user, you will see private and public
        events, otherwise you will only see public events.

        :param bool public: (optional), determines if the authenticated user
            sees both private and public or just public
        :param int number: (optional), number of events to return. Default: -1
            returns all events available
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Event <github3.events.Event>`\ s
        """
        path = ["received_events"]
        if public:
            path.append("public")
        url = self._build_url(*path, base_url=self._api)
        return self._iter(int(number), url, Event, etag=etag)

    def organizations(self, number=-1, etag=None):
        r"""Iterate over organizations the user is member of.

        :param int number: (optional), number of organizations to return.
            Default: -1 returns all available organization
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of
            :class:`ShortOrganization <github3.orgs.ShortOrganization>`\ s
        """
        # Import here, because a toplevel import causes an import loop
        from .orgs import ShortOrganization

        url = self._build_url("orgs", base_url=self._api)
        return self._iter(int(number), url, ShortOrganization, etag=etag)

    def starred_repositories(
        self, sort=None, direction=None, number=-1, etag=None
    ):
        """Iterate over repositories starred by this user.

        .. versionchanged:: 0.5
           Added sort and direction parameters (optional) as per the change in
           GitHub's API.

        :param int number: (optional), number of starred repos to return.
            Default: -1, returns all available repos
        :param str sort: (optional), either 'created' (when the star was
            created) or 'updated' (when the repository was last pushed to)
        :param str direction: (optional), either 'asc' or 'desc'. Default:
            'desc'
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`~github3.repos.repo.StarredRepository`
        """
        from .repos import Repository, StarredRepository

        params = {"sort": sort, "direction": direction}
        self._remove_none(params)
        url = self.starred_urlt.expand(owner=None, repo=None)
        return self._iter(
            int(number),
            url,
            StarredRepository,
            params,
            etag,
            headers=Repository.STAR_HEADERS,
        )

    def subscriptions(self, number=-1, etag=None):
        """Iterate over repositories subscribed to by this user.

        :param int number: (optional), number of subscriptions to return.
            Default: -1, returns all available
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Repository <github3.repos.Repository>`
        """
        from .repos import ShortRepository

        url = self._build_url("subscriptions", base_url=self._api)
        return self._iter(int(number), url, ShortRepository, etag=etag)

    @requires_auth
    def rename(self, login):
        """Rename the user.

        .. note::

            This is only available for administrators of a GitHub Enterprise
            instance.

        :param str login: (required), new name of the user
        :returns: bool
        """
        url = self._build_url("admin", "users", self.login)
        payload = {"login": login}
        resp = self._boolean(self._patch(url, data=payload), 202, 403)
        return resp

    @requires_auth
    def impersonate(self, scopes=None):
        """Obtain an impersonation token for the user.

        The retrieved token will allow impersonation of the user.
        This is only available for admins of a GitHub Enterprise instance.

        :param list scopes: (optional), areas you want this token to apply to,
            i.e., 'gist', 'user'
        :returns: :class:`Authorization <Authorization>`
        """
        url = self._build_url("admin", "users", self.login, "authorizations")
        data = {}

        if scopes:
            data["scopes"] = scopes

        json = self._json(self._post(url, data=data), 201)

        return self._instance_or_null(Authorization, json)

    @requires_auth
    def revoke_impersonation(self):
        """Revoke all impersonation tokens for the current user.

        This is only available for admins of a GitHub Enterprise instance.

        :returns: bool -- True if successful, False otherwise
        """
        url = self._build_url("admin", "users", self.login, "authorizations")

        return self._boolean(self._delete(url), 204, 403)

    @requires_auth
    def promote(self):
        """Promote a user to site administrator.

        This is only available for admins of a GitHub Enterprise instance.

        :returns: bool -- True if successful, False otherwise
        """
        url = self._build_url("site_admin", base_url=self._api)

        return self._boolean(self._put(url), 204, 403)

    @requires_auth
    def demote(self):
        """Demote a site administrator to simple user.

        You can demote any user account except your own.

        This is only available for admins of a GitHub Enterprise instance.

        :returns: bool -- True if successful, False otherwise
        """
        url = self._build_url("site_admin", base_url=self._api)

        return self._boolean(self._delete(url), 204, 403)

    @requires_auth
    def suspend(self):
        """Suspend the user.

        This is only available for admins of a GitHub Enterprise instance.

        This API is disabled if you use LDAP, check the GitHub API dos for more
        information.

        :returns: bool -- True if successful, False otherwise
        """
        url = self._build_url("suspended", base_url=self._api)

        return self._boolean(self._put(url), 204, 403)

    @requires_auth
    def unsuspend(self):
        """Unsuspend the user.

        This is only available for admins of a GitHub Enterprise instance.

        This API is disabled if you use LDAP, check the GitHub API dos for more
        information.

        :returns: bool -- True if successful, False otherwise
        """
        url = self._build_url("suspended", base_url=self._api)

        return self._boolean(self._delete(url), 204, 403)

    @requires_auth
    def delete(self):
        """Delete the user.

        Per GitHub API documentation, it is often preferable to suspend the
        user.

        .. note::

            This is only available for admins of a GitHub Enterprise instance.

        :returns: bool -- True if successful, False otherwise
        """
        url = self._build_url("admin", "users", self.login)
        return self._boolean(self._delete(url), 204, 403)


class User(_User):
    """Object for the full representation of a User.

    GitHub's API returns different amounts of information about users based
    upon how that information is retrieved. This object exists to represent
    the full amount of information returned for a specific user. For example,
    you would receive this class when calling
    :meth:`~github3.github.GitHub.user`. To provide a clear distinction
    between the types of users, github3.py uses different classes with
    different sets of attributes.

    This object no longer contains information about the currently
    authenticated user (e.g., :meth:`~github3.github.GitHub.me`).

    .. versionchanged:: 1.0.0

    This object contains all of the attributes available on
    :class:`~github3.users.ShortUser` as well as the following:

    .. attribute:: bio

        The markdown formatted User's biography

    .. attribute:: blog

        The URL of the user's blog

    .. attribute:: company

        The name or GitHub handle of the user's company

    .. attribute:: created_at

        A parsed :class:`~datetime.datetime` object representing the date the
        user was created

    .. attribute:: email

        The email address the user has on their public profile page

    .. attribute:: followers_count

        The number of followers of this user

    .. attribute:: following_count

        The number of users this user follows

    .. attribute:: hireable

        Whether or not the user has opted into GitHub jobs advertising

    .. attribute:: location

        The location specified by the user on their public profile

    .. attribute:: name

        The name specified by their user on their public profile

    .. attribute:: public_gists_count

        The number of public gists owned by this user

    .. attribute: public_repos_count

        The number of public repositories owned by this user

    .. attribute:: updated_at

        A parsed :class:`~datetime.datetime` object representing the date
        the user was last updated
    """

    class_name = "User"

    def _update_attributes(self, user):
        super()._update_attributes(user)
        self.bio = user["bio"]
        self.blog = user["blog"]
        self.company = user["company"]
        self.created_at = self._strptime(user["created_at"])
        self.email = user["email"]
        self.followers_count = user["followers"]
        self.following_count = user["following"]
        self.hireable = user["hireable"]
        self.location = user["location"]
        self.name = user["name"]
        self.public_gists_count = user["public_gists"]
        self.public_repos_count = user["public_repos"]
        self.updated_at = self._strptime(user["updated_at"])


class ShortUser(_User):
    """Object for the shortened representation of a User.

    GitHub's API returns different amounts of information about users based
    upon how that information is retrieved. Often times, when iterating over
    several users, GitHub will return less information. To provide a clear
    distinction between the types of users, github3.py uses different classes
    with different sets of attributes.

    .. versionadded:: 1.0.0


    .. attribute:: avatar_url

        The URL of the avatar (possibly from Gravatar)

    .. attribute:: events_urlt

        A URITemplate object from ``uritemplate`` that can be used to generate
        an events URL

    .. attribute:: followers_url

        A string representing the resource to retrieve a User's followers

    .. attribute:: following_urlt

        A URITemplate object from ``uritemplate`` that can be used to generate
        the URL to check if this user is following ``other_user``

    .. attribute:: gists_urlt

        A URITemplate object from ``uritemplate`` that can be used to generate
        the URL to retrieve a Gist by its id

    .. attribute:: gravatar_id

        The identifier for the user's gravatar

    .. attribute:: html_url

        The URL of the user's publicly visible profile. For example,
        ``https://github.com/sigmavirus24``

    .. attribute:: id

        The unique ID of the account

    .. attribute:: login

        The username of the user, e.g., ``sigmavirus24``

    .. attribute:: organizations_url

        A string representing the resource to retrieve the organizations to
        which a user belongs

    .. attribute:: received_events_url

        A string representing the resource to retrieve the events a user
        received

    .. attribute:: repos_url

        A string representing the resource to list a user's repositories

    .. attribute:: site_admin

        A boolean attribute indicating whether the user is a member of
        GitHub's staff

    .. attribute:: starred_urlt

        A URITemplate object from ``uritemplate`` that can be used to generate
        a URL to retrieve whether the user has starred a repository.

    .. attribute:: subscriptions_url

        A string representing the resource to list a user's subscriptions

    .. attribute:: type

        A string representing the type of User account this. In all cases
        should be "User"

    .. attribute:: url

        A string of this exact resource retrievable from GitHub's API
    """

    class_name = "ShortUser"
    _refresh_to = User


class Stargazer(_User):
    """Object representing a user that has starred a repository.

    .. versionadded:: 3.0.0

    This object contains all of the attributes available on
    :class:`~github3.users.ShortUser` as well as the following:

    .. attribute:: starred_at

        The time and date that the user starred the repository this was
        queried from.
    """

    class_name = "Stargazer"
    _refresh_to = User

    def _update_attributes(self, stargazer):
        super()._update_attributes(stargazer["user"])
        self.starred_at = self._strptime(stargazer["starred_at"])


class AuthenticatedUser(User):
    """Object to represent the currently authenticated user.

    This is returned by :meth:`~github3.github.GitHub.me`. It contains the
    extra informtation that is not returned for other users such as the
    currently authenticated user's plan and private email information.

    .. versionadded:: 1.0.0

    .. versionchanged:: 1.0.0

        The ``total_private_gists`` attribute is no longer returned by
        GitHub's API and so is removed.

    This object has all of the same attribute as the
    :class:`~github3.users.ShortUser` and :class:`~github3.users.User` objects
    as well as:

    .. attribute:: disk_usage

        The amount of repository space that has been used by you, the user

    .. attribute:: owned_private_repos_count

        The number of private repositories owned by you, the user

    .. attribute:: plan

        .. note::

            When used with a Github Enterprise instance <= 2.12.7, this
            attribute will not be returned. To handle these situations
            sensitively, the attribute will be set to ``None``.
            Repositories may still have a license associated with them
            in these cases.

        The name of the plan that you, the user, have purchased
    """

    class_name = "AuthenticatedUser"

    def _update_attributes(self, user):
        super()._update_attributes(user)
        self.disk_usage = user.get("disk_usage")
        self.owned_private_repos_count = user.get("owned_private_repos")
        self.total_private_repos_count = user.get("total_private_repos")
        self.plan = user.get("plan")
        if self.plan is not None:
            self.plan = Plan(self.plan, self)


class Collaborator(_User):
    """Object for the representation of a collaborator.

    .. versionadded:: 1.1.0

    When retrieving a repository's contributors, GitHub returns the same
    information as a :class:`~github3.users.ShortUser` with an additional
    attribute:

    .. attribute:: permissions

        Admin, push, and pull permissions of a collaborator
    """

    class_name = "Collaborator"
    _refresh_to = User

    def _update_attributes(self, user):
        super()._update_attributes(user)

        self.permissions = user["permissions"]


class Contributor(_User):
    """Object for the specialized representation of a contributor.

    .. versionadded:: 1.0.0

    .. versionchanged:: 1.1.0

        This class now refreshes to a :class:`~github3.users.User`.

        The attribute ``contributions`` was renamed to ``contributions_count``,
        the documentation already declared it as ``contributions_count``, it
        was the implementation now reflects this as well.

    When retrieving a repository's contributors, GitHub returns the same
    information as a :class:`~github3.users.ShortUser` with an additional
    attribute:

    .. attribute:: contributions_count

        The number of contributions a contributor has made to the repository

    """

    class_name = "Contributor"
    _refresh_to = User

    def _update_attributes(self, contributor):
        super()._update_attributes(contributor)
        self.contributions_count = contributor["contributions"]


UserLike = t.Union[
    ShortUser, User, AuthenticatedUser, Collaborator, Contributor, str
]



"""
This module contains the classes relating to notifications.

See also: http://developer.github.com/v3/activity/notifications/
"""
from json import dumps

from . import models


class Thread(models.GitHubCore):
    """Object representing a notification thread.

    .. versionchanged:: 1.0.0

        The ``comment``, ``thread``, and ``url`` attributes are no longer
        present because GitHub stopped returning the comment that caused
        the notification.

        The ``is_unread`` method was removed since it just returned the
        ``unread`` attribute.

    This object has the following attributes:

    .. attribute:: id

        The unique identifier for this notification across all GitHub
        notifications.

    .. attribute:: last_read_at

        A :class:`~datetime.datetime` object representing the date and time
        when the authenticated user last read this thread.

    .. attribute:: reason

        The reason the authenticated user is receiving this notification.

    .. attribute:: repository

        A :class:`~github3.repos.ShortRepository` this thread originated on.

    .. attribute:: subject

        A dictionary with the subject of the notification, for example, which
        issue, pull request, or diff this is in relation to.

    .. attribute:: unread

        A boolean attribute indicating whether this thread has been read or
        not.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` representing the date and time when this
        thread was last updated.

    See also:
    http://developer.github.com/v3/activity/notifications/
    """

    def _update_attributes(self, thread):
        from . import repos

        self._api = thread["url"]
        self.id = thread["id"]
        self.last_read_at = self._strptime(thread["last_read_at"])
        self.reason = thread["reason"]
        self.repository = repos.ShortRepository(thread["repository"], self)
        self.subject = thread["subject"]
        self.unread = thread["unread"]
        self.updated_at = self._strptime(thread["updated_at"])

    def _repr(self):
        return "<Thread [{}]>".format(self.subject.get("title"))

    def delete_subscription(self):
        """Delete subscription for this thread.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("subscription", base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    def mark(self):
        """Mark the thread as read.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._boolean(self._patch(self._api), 205, 404)

    def set_subscription(self, subscribed, ignored):
        """Set the user's subscription for this thread.

        :param bool subscribed:
            (required), determines if notifications should be received from
            this thread.
        :param bool ignored:
            (required), determines if notifications should be ignored from this
            thread.
        :returns:
            new subscription
        :rtype:
            :class:`~github3.notifications.ThreadSubscription`
        """
        url = self._build_url("subscription", base_url=self._api)
        sub = {"subscribed": subscribed, "ignored": ignored}
        json = self._json(self._put(url, data=dumps(sub)), 200)
        return self._instance_or_null(ThreadSubscription, json)

    def subscription(self):
        """Check the status of the user's subscription to this thread.

        :returns:
            the subscription for this thread
        :rtype:
            :class:`~github3.notifications.ThreadSubscription`
        """
        url = self._build_url("subscription", base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(ThreadSubscription, json)


class _Subscription(models.GitHubCore):
    """This object wraps thread and repository subscription information.

    See also:
    developer.github.com/v3/activity/notifications/#get-a-thread-subscription

    .. versionchanged:: 1.0.0

        The ``is_ignored`` and ``is_subscribed`` methods were removed. Use the
        :attr`ignored` and :attr:`subscribed` attributes instead.

    This object has the following attributes:

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        the user was subscribed to the thread.

    .. attribute:: ignored

        A boolean attribute indicating whether the user ignored this.

    .. attribute:: reason

        The reason the user is subscribed to the thread.

    .. attribute:: subscribed

        A boolean attribute indicating whether the user is subscribed or not.

    .. attribute:: thread_url

        The URL of the thread resource in the GitHub API.
    """

    class_name = "_Subscription"

    def _update_attributes(self, sub):
        self._api = sub["url"]
        self.created_at = self._strptime(sub["created_at"])
        self.ignored = sub["ignored"]
        self.reason = sub["reason"]
        self.subscribed = sub["subscribed"]

    def _repr(self):
        return f"<{self.class_name} [{self.subscribed}]>"

    def delete(self):
        """Delete this subscription.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    def set(self, subscribed, ignored):
        """Set the user's subscription for this subscription.

        :param bool subscribed:
            (required), determines if notifications should be received from
            this thread.
        :param bool ignored:
            (required), determines if notifications should be ignored from this
            thread.
        """
        sub = {"subscribed": subscribed, "ignored": ignored}
        json = self._json(self._put(self._api, data=dumps(sub)), 200)
        self._update_attributes(json)


class ThreadSubscription(_Subscription):
    """This object provides a representation of a thread subscription.

    See also:
    developer.github.com/v3/activity/notifications/#get-a-thread-subscription

    .. versionchanged:: 1.0.0

        The ``is_ignored`` and ``is_subscribed`` methods were removed. Use the
        :attr`ignored` and :attr:`subscribed` attributes instead.

    This object has the following attributes:

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        the user was subscribed to the thread.

    .. attribute:: ignored

        A boolean attribute indicating whether the user ignored this.

    .. attribute:: reason

        The reason the user is subscribed to the thread.

    .. attribute:: subscribed

        A boolean attribute indicating whether the user is subscribed or not.

    .. attribute:: thread_url

        The URL of the thread resource in the GitHub API.
    """

    class_name = "ThreadSubscription"

    def _update_subscription(self, sub):
        super()._update_attributes(sub)
        self.thread_url = sub["thread_url"]


class RepositorySubscription(_Subscription):
    """This object provides a representation of a thread subscription.

    See also:
    developer.github.com/v3/activity/notifications/#get-a-thread-subscription

    .. versionchanged:: 1.0.0

        The ``is_ignored`` and ``is_subscribed`` methods were removed. Use the
        :attr`ignored` and :attr:`subscribed` attributes instead.

    This object has the following attributes:

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        the user was subscribed to the thread.

    .. attribute:: ignored

        A boolean attribute indicating whether the user ignored this.

    .. attribute:: reason

        The reason the user is subscribed to the thread.

    .. attribute:: repository_url

        The URL of the repository resource in the GitHub API.

    .. attribute:: subscribed

        A boolean attribute indicating whether the user is subscribed or not.
    """

    class_name = "RepositorySubscription"

    def _update_subscription(self, sub):
        super()._update_attributes(sub)
        self.repository_url = sub.get("repository_url")



"""Module containing the logic for a GistComment."""
from .. import decorators
from .. import models
from .. import users


class GistComment(models.GitHubCore):
    """Representation of a comment left on a :class:`~github3.gists.Gist`.

    See also: http://developer.github.com/v3/gists/comments/

    .. versionchanged:: 1.0.0

        The ``links``, ``html_url``, and ``pull_request_url`` attributes were
        removed as none of them exist in the response from GitHub.

    This object has the following attributes:

    .. attribute:: author_association

        The comment author's (:attr:`user`) association with this gist.

    .. attribute:: body

        The markdown formatted original text written by the author.

    .. attribute:: body_html

        The HTML formatted comment body.

    .. attribute:: body_text

        The plain-text formatted comment body.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        when this comment was created.

    .. attribute:: id

        The unique identifier for this comment.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        when this comment was most recently updated.

    .. attribute:: user

        A :class:`~github3.users.ShortUser` representing the author of this
        comment.
    """

    def _update_attributes(self, comment):
        self._api = comment["url"]
        self.author_association = comment["author_association"]
        self.body = comment["body"]
        self.body_html = comment["body_html"]
        self.body_text = comment["body_text"]
        self.created_at = self._strptime(comment["created_at"])
        self.id = comment["id"]
        self.updated_at = self._strptime(comment["updated_at"])
        self.user = users.ShortUser(comment["user"], self)

    def _repr(self):
        return f"<Gist Comment [{self.user.login}]>"

    @decorators.requires_auth
    def delete(self):
        """Delete this comment from the gist.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @decorators.requires_auth
    def edit(self, body):
        """Edit this comment.

        :param str body:
            (required), new body of the comment, Markdown-formatted
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        if body:
            json = self._json(
                self._patch(self._api, json={"body": body}), 200
            )
            if json:
                self._update_attributes(json)
                return True
        return False



"""All exceptions for the github3 library."""


class GitHubException(Exception):
    """The base exception class.

    .. versionadded:: 1.0.0

        Necessary to handle pre-response exceptions

    """

    pass


class GeneratedTokenExpired(GitHubException):
    """This exception is used to prevent an expired token from being used.

    .. versionadded:: 1.2.0
    """

    pass


class AppInstallationTokenExpired(GeneratedTokenExpired):
    """This exception is used to prevent an expired token from being used.

    .. versionadded:: 1.2.0
    """

    pass


class AppTokenExpired(GeneratedTokenExpired):
    """This exception is used to prevent an expired token from being used.

    .. versionadded:: 1.2.0
    """

    pass


class MissingAppAuthentication(GitHubException):
    """Raised when user tries to call a method that requires app auth.

    .. versionadded:: 1.2.0
    """

    pass


class MissingAppBearerAuthentication(MissingAppAuthentication):
    """Raised when user tries to call a method that requires app auth.

    .. versionadded:: 1.2.0
    """

    pass


class MissingAppInstallationAuthentication(MissingAppAuthentication):
    """Raised when user tries to call a method that requires app installation.

    .. versionadded:: 1.2.0
    """

    pass


class CardHasNoContentUrl(GitHubException):
    """Raised when attempting a card has no ``content_url``.

    We use this in methods to retrieve the underlying issue or pull request
    based on the ``content_url``.

    .. versionadded:: 1.3.0
    """

    pass


class GitHubError(GitHubException):
    """The base exception class for all response-related exceptions.

    .. versionchanged:: 1.0.0

        This now inherits from :class:`~github3.exceptions.GitHubException`

    .. attribute:: response

        The response object that triggered the exception

    .. attribute:: code

        The response's status code

    .. attribute:: errors

        The list of errors (if present) returned by GitHub's API
    """

    def __init__(self, resp):
        """Initialize our exception class."""
        super().__init__(resp)
        #: Response code that triggered the error
        self.response = resp
        self.code = resp.status_code
        self.errors = []
        try:
            error = resp.json()
            #: Message associated with the error
            self.msg = error.get("message")
            #: List of errors provided by GitHub
            if error.get("errors"):
                self.errors = error.get("errors")
        except Exception:  # Amazon S3 error
            self.msg = resp.content or "[No message]"

    def __repr__(self):
        return "<{} [{}]>".format(
            self.__class__.__name__, self.msg or self.code
        )

    def __str__(self):
        return f"{self.code} {self.msg}"

    @property
    def message(self):
        """The actual message returned by the API."""
        return self.msg


class IncompleteResponse(GitHubError):
    """Exception for a response that doesn't have everything it should.

    This has the same attributes as :class:`~github3.exceptions.GitHubError`
    as well as

    .. attribute:: exception

        The original exception causing the IncompleteResponse exception
    """

    def __init__(self, json, exception):
        """Initialize our IncompleteResponse."""
        self.response = None
        self.code = None
        self.json = json
        self.errors = []
        self.exception = exception
        self.msg = (
            "The library was expecting more data in the response (%r)."
            " Either GitHub modified it's response body, or your token"
            " is not properly scoped to retrieve this information."
        ) % (exception,)


class NotRefreshable(GitHubException):
    """Exception to indicate that an object is not refreshable."""

    message_format = (
        '"{}" is not refreshable because the GitHub API does '
        "not provide a URL to retrieve its contents from."
    )

    def __init__(self, object_name):
        """Initialize our NotRefreshable exception."""
        super().__init__(self.message_format.format(object_name))


class ResponseError(GitHubError):
    """The base exception for errors stemming from GitHub responses."""

    pass


class TransportError(GitHubException):
    """Catch-all exception for errors coming from Requests.

    .. versionchanged:: 1.0.0

        Now inherits from :class:`~github3.exceptions.GitHubException`.
    """

    msg_format = "An error occurred while making a request to GitHub: {0}"

    def __init__(self, exception):
        """Initialize TransportError exception."""
        self.msg = self.msg_format.format(str(exception))
        super().__init__(self, self.msg, exception)
        self.exception = exception

    def __str__(self):
        return f"{type(self.exception)}: {self.msg}"


class ConnectionError(TransportError):
    """Exception for errors in connecting to or reading data from GitHub."""

    msg_format = "A connection-level exception occurred: {0}"


class UnexpectedResponse(ResponseError):
    """Exception class for responses that were unexpected."""

    pass


class UnprocessableResponseBody(ResponseError):
    """Exception class for response objects that cannot be handled."""

    def __init__(self, message, body):
        """Initialize UnprocessableResponseBody."""
        Exception.__init__(self, message)
        self.body = body
        self.msg = message

    def __repr__(self):
        return "<{} [{}]>".format("UnprocessableResponseBody", self.body)

    def __str__(self):
        return self.message


class BadRequest(ResponseError):
    """Exception class for 400 responses."""

    pass


class AuthenticationFailed(ResponseError):
    """Exception class for 401 responses.

    Possible reasons:

    - Need one time password (for two-factor authentication)
    - You are not authorized to access the resource
    """

    pass


class ForbiddenError(ResponseError):
    """Exception class for 403 responses.

    Possible reasons:

    - Too many requests (you've exceeded the ratelimit)
    - Too many login failures
    """

    pass


class NotFoundError(ResponseError):
    """Exception class for 404 responses."""

    pass


class MethodNotAllowed(ResponseError):
    """Exception class for 405 responses."""

    pass


class NotAcceptable(ResponseError):
    """Exception class for 406 responses."""

    pass


class Conflict(ResponseError):
    """Exception class for 409 responses.

    Possible reasons:

    - Head branch was modified (SHA sums do not match)
    """

    pass


class UnprocessableEntity(ResponseError):
    """Exception class for 422 responses."""

    pass


class ClientError(ResponseError):
    """Catch-all for 400 responses that aren't specific errors."""

    pass


class ServerError(ResponseError):
    """Exception class for 5xx responses."""

    pass


class UnavailableForLegalReasons(ResponseError):
    """Exception class for 451 responses."""

    pass


error_classes = {
    400: BadRequest,
    401: AuthenticationFailed,
    403: ForbiddenError,
    404: NotFoundError,
    405: MethodNotAllowed,
    406: NotAcceptable,
    409: Conflict,
    422: UnprocessableEntity,
    451: UnavailableForLegalReasons,
}


def error_for(response):
    """Return the appropriate initialized exception class for a response."""
    klass = error_classes.get(response.status_code)
    if klass is None:
        if 400 <= response.status_code < 500:
            klass = ClientError
        if 500 <= response.status_code < 600:
            klass = ServerError
    return klass(response)



"""
This module contains all the classes relating to Git Data.

See also: http://developer.github.com/v3/git/
"""
import base64
from json import dumps

from . import models
from .decorators import requires_auth


class Blob(models.GitHubCore):
    """This object provides an interface to the API representation of a blob.

    See also: http://developer.github.com/v3/git/blobs/

    .. versionchanged:: 1.0.0

       - The :attr:`content` is no longer forcibly coerced to bytes.

    This object has the following atributes

    .. attribute:: content

        The raw content of the blob. This may be base64 encoded text. Use
        :meth:`decode_content` to receive the non-encoded text.

    .. attribute:: encoding

        The encoding that GitHub reports for this blob's content.

    .. attribute:: size

        The size of this blob's content in bytes.

    .. attribute:: sha

        The SHA1 of this blob's content.
    """

    def _update_attributes(self, blob):
        self._api = blob["url"]
        self.content = blob["content"]
        self.encoding = blob["encoding"]
        self.size = blob["size"]
        self.sha = blob["sha"]

    def _repr(self):
        return f"<Blob [{self.sha:.10}]>"

    def decode_content(self):
        """Return the unencoded content of this blob.

        If the content is base64 encoded, this will properly decode it.
        Otherwise, it will return the content as returned by the API.

        :returns:
            Decoded content as text
        :rtype:
            unicode
        """
        if self.encoding == "base64" and self.content:
            return base64.b64decode(self.content.encode("utf-8")).decode(
                "utf-8"
            )
        return self.content


class _Commit(models.GitHubCore):
    class_name = "_Commit"

    def _update_attributes(self, commit):
        self._api = commit["url"]
        self.author = commit["author"]
        if self.author.get("name"):
            self._author_name = self.author["name"]
        self.committer = commit["committer"]
        if self.committer:
            self._commit_name = self.committer.get("name")
        self.message = commit["message"]
        self.tree = CommitTree(commit["tree"], self)

    def _repr(self):
        return f"<{self.class_name} [{self.sha}]>"


class Commit(_Commit):
    """This represents a commit as returned by the git API.

    This is distinct from :class:`~github3.repos.commit.RepoCommit`.
    Primarily this object represents the commit data stored by git and
    it has no relationship to the repository on GitHub.

    See also: http://developer.github.com/v3/git/commits/

    This object has all of the attributes of a
    :class:`~github3.git.ShortCommit` as well as the following attributes:

    .. attribute:: parents

        The list of commits that are the parents of this commit. This may be
        empty if this is the initial commit, or it may have several if it is
        the result of an octopus merge. Each parent is represented as a
        dictionary with the API URL and SHA1.

    .. attribute:: sha

        The unique SHA1 which identifies this commit.

    .. attribute:: verification

        The GPG verification data about this commit. See
        https://developer.github.com/v3/git/commits/#commit-signature-verification
        for more information.
    """

    class_name = "Commit"

    def _update_attributes(self, commit):
        super()._update_attributes(commit)
        self.parents = commit["parents"]
        self.sha = commit["sha"]
        if not self.sha:
            i = self._api.rfind("/")
            self.sha = self._api[i + 1 :]
        self._uniq = self.sha
        self.verification = commit["verification"]


class ShortCommit(_Commit):
    """This represents a commit as returned by the git API.

    This is distinct from :class:`~github3.repos.commit.RepoCommit`.
    Primarily this object represents the commit data stored by git. This
    shorter representation of a Commit is most often found on a
    :class:`~github3.repos.commit.RepoCommit` to represent the git data
    associated with it.

    See also: http://developer.github.com/v3/git/commits/

    This object has the following attributes:

    .. attribute:: author

        This is a dictionary with at least the name and email of the author
        of this commit as well as the date it was authored.

    .. attribute:: committer

        This is a dictionary with at least the name and email of the committer
        of this commit as well as the date it was committed.

    .. attribute:: message

        The commit message that describes the changes as written by the author
        and committer.

    .. attribute:: tree

        The git tree object this commit points to.
    """

    class_name = "ShortCommit"
    _refresh_to = Commit


class Reference(models.GitHubCore):
    """Object representing a git reference associated with a repository.

    This represents a reference (or ref) created on a repository via git.

    See also: http://developer.github.com/v3/git/refs/

    This object has the following attributes:

    .. attribute:: object

        A :class:`~github3.git.GitObject` that this reference points to.

    .. attribute:: ref

        The string path to the reference, e.g., ``'refs/heads/sc/feature-a'``.
    """

    def _update_attributes(self, ref):
        self._api = ref["url"]
        self.object = GitObject(ref["object"], self)
        self.ref = ref["ref"]

    def _repr(self):
        return f"<Reference [{self.ref}]>"

    @requires_auth
    def delete(self):
        """Delete this reference.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def update(self, sha, force=False):
        """Update this reference.

        :param str sha:
            (required), sha of the reference
        :param bool force:
            (optional), force the update or not
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        data = {"sha": sha, "force": force}
        json = self._json(self._patch(self._api, data=dumps(data)), 200)
        if json:
            self._update_attributes(json)
            return True
        return False


class GitObject(models.GitHubCore):
    """This object represents an arbitrary 'object' in git.

    This object is intended to be versatile and is usually found on one of the
    following:

    - :class:`~github3.git.Reference`
    - :class:`~github3.git.Tag`

    This object has the following attributes:

    .. attribute:: sha

        The SHA1 of the object this is representing.

    .. attribute:: type

        The name of the type of object this is representing.
    """

    def _update_attributes(self, obj):
        self._api = obj["url"]
        self.sha = obj["sha"]
        self.type = obj["type"]

    def _repr(self):
        return f"<Git Object [{self.sha}]>"


class Tag(models.GitHubCore):
    """This represents an annotated tag.

    Tags are a special kind of git reference and annotated tags have more
    information than lightweight tags.

    See also: http://developer.github.com/v3/git/tags/

    This object has the following attributes:

    .. attribute:: message

        This is the message that was written to accompany the creation of the
        annotated tag.

    .. attribute:: object

        A :class:`~github3.git.GitObject` that represents the underlying git
        object.

    .. attribute:: sha

        The SHA1 of this tag in the git repository.

    .. attribute:: tag

        The "lightweight" tag (or reference) that backs this annotated tag.

    .. attribute:: tagger

        The person who created this tag.
    """

    def _update_attributes(self, tag):
        self._api = tag["url"]
        self.message = tag["message"]
        self.object = GitObject(tag["object"], self)
        self.sha = tag["sha"]
        self.tag = tag["tag"]
        self.tagger = tag["tagger"]

    def _repr(self):
        return f"<Tag [{self.tag}]>"


class CommitTree(models.GitHubCore):
    """This object represents the abbreviated tree data in a commit.

    The API returns different representations of different objects. When
    representing a :class:`~github3.git.ShortCommit` or
    :class:`~github3.git.Commit`, the API returns an abbreviated
    representation of a git tree.

    This object has the following attributes:

    .. attribute:: sha

        The SHA1 of this tree in the git repository.
    """

    def _update_attributes(self, tree):
        self._api = tree["url"]
        self.sha = tree["sha"]

    def _repr(self):
        return f"<CommitTree [{self.sha}]>"

    def to_tree(self):
        """Retrieve a full Tree object for this CommitTree.

        :returns:
            The full git data about this tree
        :rtype:
            :class:`~github3.git.Tree`
        """
        json = self._json(self._get(self._api), 200)
        return self._instance_or_null(Tree, json)

    refresh = to_tree


class Tree(models.GitHubCore):
    """This represents a tree object from a git repository.

    Trees tend to represent directories and subdirectories.

    See also: http://developer.github.com/v3/git/trees/

    This object has the following attributes:

    .. attribute:: sha

        The SHA1 of this tree in the git repository.

    .. attribute:: tree

        A list that represents the nodes in the tree. If this list has members
        it will have instances of :class:`~github3.git.Hash`.
    """

    def _update_attributes(self, tree):
        self._api = tree["url"]
        self.sha = tree["sha"]
        self.tree = tree["tree"]
        if self.tree:
            self.tree = [Hash(t, self) for t in self.tree]

    def _repr(self):
        return f"<Tree [{self.sha}]>"

    def __eq__(self, other):
        return self.as_dict() == other.as_dict()

    def __ne__(self, other):
        return self.as_dict() != other.as_dict()

    def recurse(self):
        """Recurse into this tree.

        :returns:
            A new tree
        :rtype:
            :class:`~github3.git.Tree`
        """
        json = self._json(
            self._get(self._api, params={"recursive": "1"}), 200
        )
        return self._instance_or_null(Tree, json)


class Hash(models.GitHubCore):
    """This is used to represent the elements of a tree.

    This provides the path to the object and the type of object it is. For
    a brief explanation of what these types are and represent, this
    StackOverflow question answers some of that:
    https://stackoverflow.com/a/18605496/1953283

    See also: http://developer.github.com/v3/git/trees/#create-a-tree

    This object has the following attributes:

    .. attribute:: mode

        The mode of the file, directory, or link.

    .. attribute:: path

        The path to the file, directory, or link.

    .. attribute:: sha

        The SHA1 for this hash.

    .. attribute:: size

        This attribute is only not ``None`` if the :attr:`type` is not a tree.

    .. attribute:: type

        The type of git object this is representing, e.g., tree, blob, etc.
    """

    def _update_attributes(self, info):
        self._api = info.get("url")
        self.mode = info["mode"]
        self.path = info["path"]
        self.sha = info["sha"]
        self.size = None
        self.type = info["type"]

        if self.type != "tree":
            self.size = info.get("size")

    def _repr(self):
        return f"<Hash [{self.sha}]>"



"""
github3.gists
=============

Module which contains all the gist related material.

Sub-modules:
    github3.gists.gist
    github3.gists.file
    github3.gists.comment
    github3.gists.history

See also: http://developer.github.com/v3/gists/
"""
from .gist import Gist
from .gist import ShortGist

__all__ = ("Gist", "ShortGist")



"""The module that holds much of the metadata about github3.py."""
__package_name__ = "github3.py"
__title__ = "github3"
__author__ = "Ian Stapleton Cordasco"
__author_email__ = "graffatcolmingov@gmail.com"
__license__ = "Modified BSD"
__copyright__ = "Copyright 2012 Ian Stapleton Cordasco"
__version__ = "4.0.1"
__version_info__ = tuple(
    int(i) for i in __version__.split(".") if i.isdigit()
)
__url__ = "https://github3.readthedocs.io"

__all__ = (
    "__package_name__",
    "__title__",
    "__author__",
    "__author_email__",
    "__license__",
    "__copyright__",
    "__version__",
    "__version_info__",
    "__url__",
)



"""This module contains the Comparison object."""
from . import commit
from .. import models


class Comparison(models.GitHubCore):
    """A representation of a comparison between two or more commit objects.

    See also:
    http://developer.github.com/v3/repos/commits/#compare-two-commits

    This object has the following attributes::

    .. attribute:: ahead_by

        The number of commits between the head and base commit.

    .. attribute:: base_commit

        A :class:`~github3.repos.commit.ShortCommit` representing the base
        commit in this comparison.

    .. attribute:: behind_by

        The number of commits the head commit is behind the base.

    .. attribute:: original_commits

        .. versionchanged:: 4.0.0


        A list of :class:`~github3.repos.commit.ShortCommit` objects
        representing up to the first 250 commits in the comparison.

        If a comparison has more than 250 commits, use the commits function
        to iterate over all available commits.

    .. attribute:: diff_url

        The URL to retrieve the diff between the head and base commits.

    .. attribute:: files

        A list of dictionaries describing each of the modified files in the
        comparison.

    .. attribute:: html_url

        The URL to view the comparison in a browser.

    .. attribute:: patch_url

        The URL to retrieve the patch-formatted diff of this comparison.

    .. attribute:: permalink_url

        The permanent URL to retrieve this comparison.

    .. attribute:: status

        Whether the head commit is ahead or behind of base.

    .. attribute:: total_commits

        The total number of commits difference.
    """

    def _update_attributes(self, compare):
        self._api = compare["url"]
        self.ahead_by = compare["ahead_by"]
        self.base_commit = commit.ShortCommit(compare["base_commit"], self)
        self.behind_by = compare["behind_by"]
        self.total_commits = compare["total_commits"]
        self.original_commits = [
            commit.ShortCommit(com, self) for com in compare["commits"]
        ]
        self.diff_url = compare["diff_url"]
        self.files = compare["files"]
        self.html_url = compare["html_url"]
        self.patch_url = compare["patch_url"]
        self.permalink_url = compare["permalink_url"]
        self.status = compare["status"]
        self._uniq = self.original_commits

    def _repr(self):
        return f"<Comparison of {self.total_commits} commits>"

    def diff(self):
        """Retrieve the diff for this comparison.

        :returns:
            the diff as a bytes object
        :rtype:
            bytes
        """
        resp = self._get(
            self._api, headers={"Accept": "application/vnd.github.diff"}
        )
        return resp.content if self._boolean(resp, 200, 404) else b""

    def patch(self):
        """Retrieve the patch formatted diff for this commit.

        :returns:
            the patch as a bytes object
        :rtype:
            bytes
        """
        resp = self._get(
            self._api, headers={"Accept": "application/vnd.github.patch"}
        )
        return resp.content if self._boolean(resp, 200, 404) else b""

    def commits(self, number=-1, etag=None):
        """Iterate over the commits available for this comparison.

        :param int number:
            (optional), Number of assets to return
        :param str etag:
            (optional), last ETag header sent
        :returns:
            generator of asset objects
        :rtype:
            :class:`~github3.repos.commit.ShortCommit`
        """
        return self._iter(
            number,
            self._api,
            commit.ShortCommit,
            list_key="commits",
            etag=etag,
        )



"""Repository and contributor stats logic."""
import datetime

import dateutil.tz

from .. import models
from .. import users


def alternate_week(week):
    """Map GitHub 'short' data to usable data.

    .. note:: This is not meant for public consumption

    :param dict week:
        The week's statistical data from GitHub
    :returns:
        Huamnized week statistical data
    :rtype:
        dict
    """
    start_of_week = datetime.datetime.utcfromtimestamp(int(week["w"]))
    return {
        "start of week": start_of_week.replace(tzinfo=dateutil.tz.tzutc()),
        "additions": week["a"],
        "deletions": week["d"],
        "commits": week["c"],
    }


class ContributorStats(models.GitHubCore):
    """Representation of a user's contributor statistics to a repository.

    See also http://developer.github.com/v3/repos/statistics/

    This object has the following attributes:

    .. attribute:: author

        A :class:`~github3.users.ShortUser` representing the contributor
        whose stats this object represents.

    .. attribute:: total

        The total number of commits authored by :attr:`author`.

    .. attribute:: weeks

        A list of dictionaries containing weekly statistical data.

    .. attribute:: alternate_weeks

        .. note::

            :mod:`github3` generates this data for a more humane interface
            to the data in :attr:`weeks`.

        A list of dictionaries that provide an easier to remember set of
        keys as well as a :class:`~datetime.datetime` object representing the
        start of the week. The dictionary looks vaguely like:

        .. code-block:: python

            {
                'start of week': datetime(2013, 5, 5, 5, 0, tzinfo=tzutc())
                'additions': 100,
                'deletions': 150,
                'commits': 5,
            }

    """

    def _update_attributes(self, stats_object):
        self.author = users.ShortUser(stats_object["author"], self)
        self.total = stats_object["total"]
        self.weeks = stats_object["weeks"]
        alt_weeks = self.weeks
        if alt_weeks:
            alt_weeks = [alternate_week(w) for w in self.weeks]
        self.alternate_weeks = self.alt_weeks = alt_weeks

    def _repr(self):
        return f"<Contributor Statistics [{self.author}]>"
